/*
FILE:     CifFileObjBase.C
*/
/*
VERSION:  7.001
*/
/*
DATE:     5/18/2006
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2006 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
/* 
  PURPOSE:    Base class for read/write cif files
*/

#include "GenString.h"
#include "CifFileObjBase.h"


CifFileObjBase::CifFileObjBase(const string& objFileName, eFileMode fileMode, int maxLineLength, const string& nullString, int verbose, const StringCompare::eCompareType caseSense) : TableFile(caseSense) {
  Clear();
  _verbose = verbose;
  _maxCifLineLength = maxLineLength;

  _nullString = nullString;

  OpenFile(objFileName, fileMode);
}

void CifFileObjBase::Clear() {
  _nullString.clear();
  _cifIn = NULL;
  _verbose=0;
}

void CifFileObjBase::Reset() {
  _nullString.clear();
}




int CifFileObjBase::Write(const string& cifFileName, int sortFlag, int emptyTable) {
  int iret=0;
  ofstream cifo(cifFileName.c_str(),ios::out|ios::trunc);
  iret =Write(cifo, sortFlag, emptyTable);
  cifo.close();
  return (iret);
}


int CifFileObjBase::Write(ostream &cifo, int sortFlag, int emptyTable) {
  int iret=0;
  if (sortFlag)
    iret = WriteSort(cifo, emptyTable);
  else
    iret = WriteUnSort(cifo, emptyTable);
  return (iret);
}


int CifFileObjBase::WriteSort(ostream &cifo, int emptyTable) {
  ISTable *tblP = NULL;
  
  vector<string> colNames;
  vector<string> rowValues;
  
  int ib0, jt, i, j, k, l, linePos, numColumn, numRow, ilen;
  int cwid, *cwidth=NULL;
  string tid;
 
  string loopBlockName;
  string loopTableName;
  string loopBlockTableName;
  for (ib0=0; ib0 < (int)(_blocks.size()); ib0++) {

    loopBlockName = _blocks[ib0].GetName();
    
    cifo << "data_" << loopBlockName << endl;

    tid=_BuildTableId(loopBlockName, " ");
    tid.clear();


   // Copy into a new array/sorted
   vector<Table> sorted = _tables;

   sort(sorted.begin(), sorted.end());

   for (jt=0; jt < (int)(sorted.size()); jt++)
   {
    loopTableName = sorted[jt].GetName();
    loopBlockTableName = sorted[jt].GetBlock();
// VLAD - EXAMINE See this same loop in DICFileObjBase or DDLFileObjBase
// VLAD PERFORMANCE. The old BTreeObj code was doing a partial search
// on tids. It searched for "_0_ " in the tree. Even though the exact
// search was never found, it moved the pointer down the tree and reached
// the appropriate "_0_ blah" for example and then processed from that point
// onward the entries that are exact match. This kind of modified proprietory
// search is better than STL. We need to represent this subsearch using
// different STL structures to be able to have better performance than what
// we currently have.
      if (loopBlockTableName != loopBlockName)
        continue;  
      tblP = GetTablePtr(loopBlockTableName, loopTableName);
      numColumn = tblP->GetNumColumns();
      numRow    = tblP->GetNumRows();
      cifo << "# " << endl;
      if (numRow <= 0) {
	if (emptyTable) {
	  tblP->GetColumnNames(colNames);
	  cwid = -1;
	  for (i=0; i< numColumn; i++) {
	    ilen = colNames[i].size();
	    if (  ilen > cwid) cwid=ilen;
	  }
	
	  for (i=0; i< numColumn; i++) {
	    _PrintItemName(cifo, loopTableName, colNames[i], linePos);
	    ilen = colNames[i].size();
	    for (k=0; k < 2+cwid-ilen; k++) cifo << " ";
	    linePos = cwid+2;
	  
	    cifo << _nullString; 
	    if (linePos != 0) cifo << endl;
	  }
	  colNames.clear();
	}
      }
      else if (numRow == 1) {
	tblP->GetColumnNames(colNames);
	tblP->GetRow(rowValues, 0);
	cwid = -1;
	for (i=0; i< numColumn; i++) {
	  ilen = colNames[i].size();
	  if (  ilen > cwid) cwid=ilen;
	}
	
	for (i=0; i< numColumn; i++) {
	  _PrintItemName(cifo, loopTableName, colNames[i], linePos);
	  ilen = colNames[i].size();
	  for (k=0; k < 2+cwid-ilen; k++) cifo << " ";
	  linePos = cwid+2;
	  
	  _PrintItemValue(cifo, rowValues[i], linePos);
	  if (linePos != 0) cifo << endl;
	}
	colNames.clear();
	rowValues.clear();
	
      } 
      else {
	
	cifo << "loop_" << endl;
	
	tblP->GetColumnNames(colNames);
	for (i=0; i< numColumn; i++) {
	  _PrintItemName(cifo, loopTableName, colNames[i], linePos);
	  cifo << endl;
	}
	
	cwidth  = new int[numColumn];
	
	for (i=0; i< numColumn; i++) { cwidth[i]=-1; }
	j = 0;
	l = 0;
	while (j < numRow) {
	  tblP->GetRow(rowValues, l);
	  if (!rowValues.empty()) {
	    for (i=0; i< numColumn; i++) {
	      ilen = rowValues[i].size();
	      if ( _IsQuotableText(rowValues[i])) ilen+=2;
	      
	      if (  ilen > cwidth[i]) cwidth[i]=ilen;
	    }
	    rowValues.clear();
	    j++;
	  }
	  l++;
	}
	j = 0;
	l = 0;
	while (j < numRow) {
	  linePos = 0;
	  tblP->GetRow(rowValues, l);
	  if (!rowValues.empty()) {
	    for (i=0; i< numColumn; i++) {
	      
	      ilen=_PrintItemValue(cifo, rowValues[i], linePos); 
	      if (linePos != 0) {
		if (linePos + cwidth[i]-ilen < _maxCifLineLength)
		  for (k=0; k < cwidth[i]-ilen; k++) cifo << " ";
		linePos +=    cwidth[i]-ilen;
	      }
	    }
	    if (linePos != 0) cifo << endl;
	    rowValues.clear();
	    j++;
	  }
	  l++;
	}
	if (cwidth) delete[] cwidth;
	colNames.clear();
      }
    }
    cifo <<"# " << endl;
  }
  return(1);
}



int CifFileObjBase::WriteUnSort(ostream &cifo, int emptyTable) {

ISTable *tblP = NULL;
  
  vector<string> colNames;
  vector<string> rowValues;
  
  int ib0, jt, i, j, k, l, linePos, numColumn, numRow, ilen;
  int cwid, *cwidth=NULL;
  int irow;
  
  string loopBlockName;
  string loopTableName;
  string loopBlockTableName;
  for (ib0=0; ib0 < (int)(_blocks.size()); ib0++) {

    loopBlockName = _blocks[ib0].GetName();
#if DEBUG    
    cerr <<  "Writing data block " << loopBlockName << "["<< ib0 << " of " 
	 << _blocks.size() << "]"<< " Null is " << _nullString << endl;
#endif
    
    cifo << "data_" << loopBlockName << endl;
   for (jt=0; jt < (int)(_tables.size()); jt++)
   {
    loopTableName = _tables[jt].GetName();
    loopBlockTableName = _tables[jt].GetBlock();

      if (loopBlockTableName != loopBlockName)
        continue;
	
	tblP = GetTablePtr(loopBlockTableName,loopTableName);
#ifdef VLAD_DEBUG
        if (tblP->GetName() == string("item"))
        {
            int a = 1;
            a++;
            // tblP->PrintTable();
        }
#endif
	numColumn = tblP->GetNumColumns();
	numRow    = tblP->GetNumRows();
#if DEBUG
	cerr << "Table " << loopTableName << " Rows  " << numRow 
	     << " Columns " << numColumn << endl;
#endif
	cifo << "# " << endl;
	if (numRow <= 0) {
	  if (emptyTable) {
	    tblP->GetColumnNames(colNames);
	    cwid = -1;
	    for (i=0; i< numColumn; i++) {
	      ilen = colNames[i].size();
	      if (  ilen > cwid) cwid=ilen;
	    }
	    
	    for (i=0; i< numColumn; i++) {
	      _PrintItemName(cifo, loopTableName, colNames[i], linePos);
	      ilen = colNames[i].size();
	      for (k=0; k < 2+cwid-ilen; k++) cifo << " ";
	      linePos = cwid+2;
	      
	      cifo << _nullString; 
	      if (linePos != 0) cifo << endl;
	    }
	    colNames.clear();
	  }
	}
	else if (numRow == 1) {
	  tblP->GetColumnNames(colNames);
	  irow = 0;
	  tblP->GetRow(rowValues, irow);
	  cwid = -1;
	  for (i=0; i< numColumn; i++) {
	    ilen = colNames[i].size();
	    if (  ilen > cwid) cwid=ilen;
	  }
	  
	  for (i=0; i< numColumn; i++) {
	    _PrintItemName(cifo, loopTableName, colNames[i], linePos);
	    ilen = colNames[i].size();
	    for (k=0; k < 2+cwid-ilen; k++) cifo << " ";
	    linePos = cwid+2;
	    
	    _PrintItemValue(cifo, rowValues[i], linePos);
	    if (linePos != 0) cifo << endl;
	  }
	  colNames.clear();
	  rowValues.clear();
	  
	} else {
	  
	  cifo << "loop_" << endl;
	  
	  tblP->GetColumnNames(colNames);
	  for (i=0; i< numColumn; i++) {
	    _PrintItemName(cifo, loopTableName, colNames[i], linePos);
	    cifo << endl;
	  }
	  
	  cwidth  = new int[numColumn];
	  
	  for (i=0; i< numColumn; i++) { cwidth[i]=-1; }
	  j = 0;
	  l = 0;
	  while (j < numRow) {
	    tblP->GetRow(rowValues, l);
	    if (!rowValues.empty()) {
	      for (i=0; i< numColumn; i++) {
		ilen = rowValues[i].size();
		if ( _IsQuotableText(rowValues[i])) ilen+=2;
		
		if (  ilen > cwidth[i]) cwidth[i]=ilen;
	      }
	      rowValues.clear();
	      j++;
	    }
	    l++;
	  }
#if DEBUG1
	  for (i=0; i< numColumn; i++) {
	    cerr << "Col " << i << " width " << cwidth[i] << endl;
	  }
#endif
	  j = 0;
	  l = 0;
	  while (j < numRow) {
	    linePos = 0;
	    tblP->GetRow(rowValues, l);
	    if (!rowValues.empty()) {
	      for (i=0; i< numColumn; i++) {
	      
		ilen=_PrintItemValue(cifo, rowValues[i], linePos); 
#if DEBUG1
		cerr << "Row " << j <<  " Col " << i <<  " ilen " << ilen 
		     << " Length " << rowValues[i].size() 
		   << " linePos " << linePos 
		     << " Value ["<< rowValues[i] << "]" <<endl;
#endif
		if (linePos != 0) {
		  if (linePos + cwidth[i]-ilen < _maxCifLineLength)
		    for (k=0; k < cwidth[i]-ilen; k++) cifo << " ";
		  linePos +=    cwidth[i]-ilen;
		}
	      }
	      if (linePos != 0) cifo << endl;
	      rowValues.clear();
	      j++;
	    }
	    l++;
	  }
	  if (cwidth) delete[] cwidth;
	  colNames.clear();
	}
    }
    cifo <<"# " << endl;
  }
  return(1);
}


  
int CifFileObjBase::Write(const string& cifFileName, vector<string>& catOrder, int emptyTable) {
  int iret=0;
  ofstream cifo(cifFileName.c_str(),ios::out|ios::trunc);
  iret =Write(cifo, catOrder, emptyTable);
  cifo.close();
  return (iret);
}

int CifFileObjBase::Write(ostream &cifo, vector<string>& catOrder,int emptyTable) {
  ISTable *tblP = NULL;
  
  vector<string> colNames;
  vector<string> rowValues;
  
  int ib0, jt, i, j, k, m, linePos, numColumn, numRow, ilen;
  int irow;
  int cwid, *cwidth=NULL;
  string tid;

  int orderLen = catOrder.size();
  vector<int> order(_tables.size());
  int found;
  
  string loopBlockName;
  string loopTableNameT;
  string loopTableBlockNameT;
  string loopTableName;
  string loopTableBlockName;
  for (ib0=0; ib0 < (int)(_blocks.size()); ib0++) {
    loopBlockName = _blocks[ib0].GetName();
    
    cifo << "data_" << loopBlockName << endl;
    
    for (int l=0; l<orderLen; l++) {
      tid=_BuildTableId(loopBlockName, catOrder[l]);

      jt = -1;
      found = -1;
      vector<Table>::iterator pos = find(_tables.begin(),
        _tables.end(), tid);
      if (pos != _tables.end())
      {
          jt = pos - _tables.begin();
          found = jt;
      }

      tid.clear();
      if (found>=0) {
	order[jt] = 1;

        loopTableName = _tables[jt].GetName();
        loopTableBlockName = _tables[jt].GetBlock();

//**********************
      tblP = GetTablePtr(loopTableBlockName,loopTableName);
      numColumn = tblP->GetNumColumns();
      numRow    = tblP->GetNumRows();
      cifo << "# " << endl;
      if (numRow <= 0) {
	if (emptyTable) {
	  tblP->GetColumnNames(colNames);
	  cwid = -1;
	  for (i=0; i< numColumn; i++) {
	    ilen = colNames[i].size();
	    if (  ilen > cwid) cwid=ilen;
	  }
	
	  for (i=0; i< numColumn; i++) {
	    _PrintItemName(cifo, loopTableName, colNames[i], linePos);
	    ilen = colNames[i].size();
	    for (k=0; k < 2+cwid-ilen; k++) cifo << " ";
	    linePos = cwid+2;
	  
	    cifo << _nullString; 
	    if (linePos != 0) cifo << endl;
	  }
	  colNames.clear();
	}
      }
      else if (numRow == 1) {
	tblP->GetColumnNames(colNames);
	  irow = 0;
	  tblP->GetRow(rowValues, irow);
	cwid = -1;
	for (i=0; i< numColumn; i++) {
	  ilen = colNames[i].size();
	  if (  ilen > cwid) cwid=ilen;
	}
	
	for (i=0; i< numColumn; i++) {
	  _PrintItemName(cifo, loopTableName, colNames[i], linePos);
	  ilen = colNames[i].size();
	  for (k=0; k < 2+cwid-ilen; k++) cifo << " ";
	  linePos = cwid+2;
	  
	  _PrintItemValue(cifo, rowValues[i], linePos);
	  if (linePos != 0) cifo << endl;
	}
	colNames.clear();
	rowValues.clear();
	
      } 
      else {
	
	cifo << "loop_" << endl;
	
	tblP->GetColumnNames(colNames);
	for (i=0; i< numColumn; i++) {
	  _PrintItemName(cifo, loopTableName, colNames[i], linePos);
	  cifo << endl;
	}
	
	cwidth  = new int[numColumn];
	
	for (i=0; i< numColumn; i++) { cwidth[i]=-1; }
	j = 0;
	m = 0;
	while (j < numRow) {
	  tblP->GetRow(rowValues, m);
	  if (!rowValues.empty()) {
	    for (i=0; i< numColumn; i++) {
	      ilen = rowValues[i].size();
	      if ( _IsQuotableText(rowValues[i])) ilen+=2;
	      
	      if (  ilen > cwidth[i]) cwidth[i]=ilen;
	    }
	    rowValues.clear();
	    j++;
	  }
	  m++;
	}
	j = 0;
	m = 0;
	while (j < numRow) {
	  linePos = 0;
	  tblP->GetRow(rowValues, m);
	  if (!rowValues.empty()) {
	    for (i=0; i< numColumn; i++) {
	      
	      ilen=_PrintItemValue(cifo, rowValues[i], linePos); 
	      if (linePos != 0) {
		if (linePos + cwidth[i]-ilen < _maxCifLineLength)
		  for (k=0; k < cwidth[i]-ilen; k++) cifo << " ";
		linePos +=    cwidth[i]-ilen;
	      }
	    }
	    if (linePos != 0) cifo << endl;
	    rowValues.clear();
	    j++;
	  }
	  m++;
	}
	if (cwidth) delete[] cwidth;
	colNames.clear();
      }
//**************************	
      }
    }
    tid=_BuildTableId(loopBlockName, " ");
    tid.clear();
   for (int jtt=0; jtt < (int)(_tables.size()); jtt++)
   {
    loopTableNameT = _tables[jtt].GetName();
    loopTableBlockNameT = _tables[jtt].GetBlock();
      if (loopTableBlockNameT != loopBlockName)
        continue;
      if (order[jtt]==0){
      tblP = GetTablePtr(loopTableBlockNameT,loopTableNameT);
      numColumn = tblP->GetNumColumns();
      numRow    = tblP->GetNumRows();
      cifo << "# " << endl;
      if (numRow <= 0) {
	if (emptyTable) {
	  tblP->GetColumnNames(colNames);
	  cwid = -1;
	  for (i=0; i< numColumn; i++) {
	    ilen = colNames[i].size();
	    if (  ilen > cwid) cwid=ilen;
	  }
	
	  for (i=0; i< numColumn; i++) {
	    _PrintItemName(cifo, loopTableNameT, colNames[i], linePos);
	    ilen = colNames[i].size();
	    for (k=0; k < 2+cwid-ilen; k++) cifo << " ";
	    linePos = cwid+2;
	  
	    cifo << _nullString; 
	    if (linePos != 0) cifo << endl;
	  }
	  colNames.clear();
	}
      }
      else if (numRow == 1) {
	tblP->GetColumnNames(colNames);
	  irow = 0;
	  tblP->GetRow(rowValues, irow);
	cwid = -1;
	for (i=0; i< numColumn; i++) {
	  ilen = colNames[i].size();
	  if (  ilen > cwid) cwid=ilen;
	}
	
	for (i=0; i< numColumn; i++) {
	  _PrintItemName(cifo, loopTableNameT, colNames[i], linePos);
	  ilen = colNames[i].size();
	  for (k=0; k < 2+cwid-ilen; k++) cifo << " ";
	  linePos = cwid+2;
	  
	  _PrintItemValue(cifo, rowValues[i], linePos);
	  if (linePos != 0) cifo << endl;
	}
	colNames.clear();
	rowValues.clear();
	
      } 
      else {
	
	cifo << "loop_" << endl;
	
	tblP->GetColumnNames(colNames);
	for (i=0; i< numColumn; i++) {
	  _PrintItemName(cifo, loopTableNameT, colNames[i], linePos);
	  cifo << endl;
	}
	
	cwidth  = new int[numColumn];
	
	for (i=0; i< numColumn; i++) { cwidth[i]=-1; }
	j = 0;
	m = 0;
	while (j < numRow) {
	  tblP->GetRow(rowValues, m);
	  if (!rowValues.empty()) {
	    for (i=0; i< numColumn; i++) {
	      ilen = rowValues[i].size();
	      if ( _IsQuotableText(rowValues[i])) ilen+=2;
	      
	      if (  ilen > cwidth[i]) cwidth[i]=ilen;
	    }
	    rowValues.clear();
	    j++;
	  }
	  m++;
	}
	j = 0;
	m = 0;
	while (j < numRow) {
	  linePos = 0;
	  tblP->GetRow(rowValues, m);
	  if (!rowValues.empty()) {
	    for (i=0; i< numColumn; i++) {
	      
	      ilen=_PrintItemValue(cifo, rowValues[i], linePos); 
	      if (linePos != 0) {
		if (linePos + cwidth[i]-ilen < _maxCifLineLength)
		for (k=0; k < cwidth[i]-ilen; k++) cifo << " ";
		linePos +=    cwidth[i]-ilen;
	      }
	    }
	    if (linePos != 0) cifo << endl;
	    rowValues.clear();
	    j++;
	  }
	  m++;
	}
	if (cwidth) delete[] cwidth;
	colNames.clear();
      }
      }
    }
    cifo <<"# " << endl;
  }
  return(1);
}





void CifFileObjBase::_PrintItemName(ostream& cifo, const string& category,
  const string& keyword, int& linePos)
{
  cifo << "_" << category << "." << keyword << " ";
  linePos = keyword.size() + category.size() + 3;
}

void CifFileObjBase::_PrintItemNameTab(ostream& cifo, const string& category,
  const string& keyword, int& linePos)
{
  int i;
  int start = 36;
  cifo << "    ";
  cifo << "_" << category << "." << keyword << " ";
  linePos = 4 + keyword.size() + category.size() + 3;
  while (linePos>start)
    start = start + 10;
  for (i=0; i<start-linePos; i++)
    cifo<<" ";
  linePos=start;
}

int CifFileObjBase::_IsQuotableText(const string& itemValue) 
{
  if (itemValue.empty() || (itemValue.size() < 1)) return 0;

  if (itemValue[0] == '_')   return 1;

  for (int i=0; i< (int)(itemValue.size()); i++) {
    if (itemValue[i] == ' ') 
      return 1;
    else if (itemValue[i] == '\n') 
      return 1;
    else if (itemValue[i] == '\'' || itemValue[i] == '\"') 
      return 1;

  }
  if (String::IsCiEqual(itemValue.substr(0, 5), "data_")) return 1;
  if (String::IsCiEqual(itemValue.substr(0, 5), "loop_")) return 1;
  if (String::IsCiEqual(itemValue.substr(0, 5), "save_")) return 1;
  return 0;
}


int CifFileObjBase::_PrintItemValue(ostream& cifo, const string& itemValue,
  int& linePos) 
{
  int i, str_len, multipleLine, multipleWord, embeddedQuotes;
  
  /* NULL VALUES */
  if (itemValue.empty() || (itemValue.size() == 0)) { 
    if (linePos + 2  < _maxCifLineLength) {
      linePos += 2;
      cifo << _nullString << " ";
    } else {
      cifo << endl << _nullString << " ";
      linePos = 2;
    }
    return 1;
  }
    
  str_len = itemValue.size();
  multipleLine = false;
  multipleWord = false;
  embeddedQuotes = false;


  for (i=0; i< str_len; i++) {
    if (itemValue[i] == ' ') 
      multipleWord = true;
    else if (itemValue[i] == '\n') 
      multipleLine =true;
    else if (itemValue[i] == '\'' || itemValue[i] == '\"') 
      embeddedQuotes =true;
    
  }
  if (itemValue[0] == '_')    multipleWord = true;
  if (itemValue[0] == ';')    multipleWord = true;

  if (String::IsCiEqual(itemValue.substr(0, 5), "data_"))
      multipleWord = true;
  if (String::IsCiEqual(itemValue.substr(0, 5), "loop_"))
      multipleWord = true;
  if (String::IsCiEqual(itemValue.substr(0, 5), "save_"))
      multipleWord = true;

  
  if (embeddedQuotes && multipleWord) multipleLine = true;
  if (embeddedQuotes)                 multipleWord = true;

  
  if (str_len >= _maxCifLineLength || multipleLine) {
    if (linePos != 0)  cifo << endl;
    cifo << ";" << itemValue;
    cifo << endl;
    cifo << ";" << endl;
    linePos = 0;
  }  else {
    if ((!multipleWord && str_len + 2 + linePos    > _maxCifLineLength) ||
	 (multipleWord && str_len + 4 + linePos    > _maxCifLineLength))  {
      cifo << endl; linePos = 0;
    }
    
    if (multipleWord) {
      cifo << "\'" << itemValue << "\' ";
      linePos += (str_len + 3);
    } else {
      cifo << itemValue << " ";
      linePos += str_len + 1;
    }
  }
  
  if (multipleWord || multipleLine || embeddedQuotes) 
    return (str_len+2);
  else
    return (str_len);
  
}


int CifFileObjBase::_PrintItemValueTab(ostream& cifo, const string& itemValue,
  int& linePos,int width) 
{
  int i, str_len, multipleLine, multipleWord, embeddedQuotes;
  int startcol=10;
  int N=0;

  if (linePos==0) {
    cifo << "          ";
    linePos = startcol;
  }
  width+=2;
  
  /* NULL VALUES */
  if (itemValue.empty() || (itemValue.size() == 0)) { 
    if (linePos + 2  < _maxCifLineLength) {
      linePos += 2;
      N += 2;
      cifo << _nullString << " ";
    } else {
      cifo << endl << _nullString << " ";
      linePos = 2;
      N = 2;
    }
    return 1;
  }
  
  str_len = itemValue.size();
  multipleLine = false;
  multipleWord = false;
  embeddedQuotes = false;


  for (i=0; i< str_len; i++) {
    if (itemValue[i] == ' ') 
      multipleWord = true;
    else if (itemValue[i] == '\n') 
      multipleLine =true;
    else if (itemValue[i] == '\'' || itemValue[i] == '\"') 
      embeddedQuotes =true;
  }
  if (itemValue[0] == '_')    multipleWord = true;
  if (itemValue[0] == ';')    multipleWord = true;

  if (String::IsCiEqual(itemValue.substr(0, 5), "data_"))
      multipleWord = true;
  if (String::IsCiEqual(itemValue.substr(0, 5), "loop_"))
      multipleWord = true;
  if (String::IsCiEqual(itemValue.substr(0, 5), "save_"))
      multipleWord = true;

  
  if (embeddedQuotes && multipleWord) multipleLine = true;
  if (embeddedQuotes)                 multipleWord = true;

  
  if (str_len >= _maxCifLineLength || multipleLine) {
    if (linePos != 0)  cifo << endl;
    cifo << ";" << itemValue;
    cifo << endl;
    cifo << ";" << endl;
    linePos = 0;
  }  else {

    if ((!multipleWord && str_len + 2 + linePos    > _maxCifLineLength) ||
	 (multipleWord && str_len + 4 + linePos    > _maxCifLineLength))  {
      cifo << endl; linePos = 0;
  if (linePos==0) {
    cifo << "          ";
    linePos = startcol;
  }
    }

    
    if (multipleWord) {
      cifo << "\'" << itemValue << "\' ";
      linePos += (str_len + 3);
      N += (str_len + 3);
    } else {
      cifo << itemValue << " ";
      linePos += str_len + 1;
      N += str_len + 1;
    }
  }
  
  if(linePos>startcol) {
    linePos=linePos+width-N;
    for (i=0; i<width-N; i++)
      cifo << " ";
  }

  if (multipleWord || multipleLine || embeddedQuotes) 
    return (str_len+2);
  else
    return (str_len);
  
}


int CifFileObjBase:: GetKeywordFromItemName(string& keyword,
  const string& itemName) {
/* ----------------------------------------------------------------------
   Purpose: CifFileObjBase::GetKeywordFromItemName(keyword, itemName) 
   
            Get the keyword part of an item name (ie. _<category>.<keyword>)
	    Return 1 for success or 0 otherwise.

 * ---------------------------------------------------------------------- */

  int i,k,ilen;


  keyword.clear();
  keyword.reserve(itemName.size());

  if (itemName.empty() ||   (ilen=itemName.size()) < 3 ) return(0);

  for (k=1; k < ilen; k++) {
    if (itemName[k] == '.') break;
  }
  if ((k==1) || (k == ilen)) return(0);

  for (i=k+1; i < ilen; i++) {
    keyword.push_back(itemName[i]);
  }
  return(1);

}


int CifFileObjBase::GetCategoryNameFromItemName(string& categoryName,
  const string& itemName) 
/* ----------------------------------------------------------------------
   Purpose: CifFileObjBase::GetCategoryNameFromItemName(categoryName, itemName) 
   
            Get the category part of an item name (ie. _<category>.<keyword>)
	    Return 1 for success or 0 otherwise.

 * ---------------------------------------------------------------------- */
{

  string::size_type dotIndex;

  categoryName.clear();

  if (itemName.empty())
  {
    return(0);
  }
 
  // Skip the first char and search for a dot
  dotIndex = itemName.find('.', 1);

  if (dotIndex == string::npos)
  {
    // If dot char not found, set its index to be the end of the string.
    dotIndex = itemName.size();
  }

  // Copy from the first char to the character prior to the dot char.
  categoryName.assign(itemName, 1, dotIndex - 1);

  return(1);
 
}


void CifFileObjBase::GetItemNames(vector<string>& itemNames,
  ISTable& isTable)
{

    vector<string> colNames;
    string cs;


    string prefix = "_" + isTable.GetName() + ".";

    isTable.GetColumnNames(colNames);

    itemNames.reserve(colNames.size());

    for (unsigned int i = 0; i < colNames.size(); i++)
    {
        cs = prefix;
        cs += colNames[i];

        itemNames.push_back(cs);
    }

}


/******* validation ******/

int CifFileObjBase::MandatoryCategoryChecking(CifFileObjBase *in, CifFileObjBase *ref,ofstream &log){
  int ret = 1;
  int i,j;
  vector<string> BlockNames;
  int numBlocks;
  ISTable * Category;
  vector<string> list;
  vector<string> target;
  vector<unsigned int> OutList;
  int mandatoryIndex;
  int errCode;
  vector<string> refBlockNames;
  int refnumBlocks;
  string cell;



  ref->GetBlockNames(refBlockNames);
  refnumBlocks = refBlockNames.size();
  if (refnumBlocks>1) log<<"WARNING - dictionary has "<<refnumBlocks<<" datablocks"<<endl;
  Category = ref->GetTablePtr(refBlockNames[0],"category");
  list.clear();
  list.push_back("mandatory_code");
  Category->CreateIndex("index0",list);
  target.clear();
  target.push_back("yes");
  Category->Search(target,list,OutList,errCode);
  if (errCode>=0) {
    mandatoryIndex = Category->GetColumnIndex("id");
    
    in->GetBlockNames(BlockNames);
    numBlocks = BlockNames.size();
    
    for (i=0; i<(int)(OutList.size()); i++) {
     //do for all tables that are mandatory
      Category->GetCell(cell,mandatoryIndex,OutList[i]);
      for (j=0; j<numBlocks; j++) {
      //do it for all datablocks
	if (!in->IsTablePresent(BlockNames[j],cell)) {
	  log<<"ERROR - table "<<cell<<" is mandatory, but is missing"<<endl;
	  ret = 0;
	}
      }
    }
    BlockNames.clear();
  }
  refBlockNames.clear();
  return ret;
}

// This method converts the string with embedded escape sequences into
// the real string. For example, a two character combination "\t" is
// converted into one character representing horizontal tab.
void CifFileObjBase::ConvertEscapedString(const string& inString,
  string& outString)
{

    outString.clear();

    outString.reserve(inString.size());

    // Go through the characters of the input string
    for (unsigned int posI = 0; posI < inString.size(); )
    {
        if (inString[posI] == '\\')
        {
            if (inString[posI + 1] == '\\')
            {
                  outString.push_back('\\');
                  posI += 2;
            }
            else if (inString[posI + 1] == 't')
            {
                  outString.push_back('\t');
                  posI += 2;
            }
            else if (inString[posI + 1] == 'n')
            {
                  outString.push_back('\n');
                  posI += 2;
            }
            else
            {
                  outString.push_back(inString[posI]);
                  posI++;
            }
        } // Current character is '\'
        else
        {
            outString.push_back(inString[posI]);
            posI++;
        } // Curent character is not '\'
    } // for (characters in the input string)

} // CifFileObjBase::ConvertEscapedString()



int CifFileObjBase::ItemsChecking(CifFileObjBase *in, CifFileObjBase *ref,ofstream &log) {
  int ret = 1;
  int i,j,k,l,m;
  vector<string> BlockNames;
  int numBlocks;
  vector<string> TableNames;
  int numTables;
  ISTable * Category;
  vector<string> ItemNames;
  ISTable * ParentTbl = NULL;
  vector<string> ParentList;  vector<string> ParentTarget;
  vector<unsigned int> OutList;
  int numRows;

  ISTable * ParentChildTbl;
  string ParentTblName;
  string keyword;
  int ParentIndex;
  string ParentName;
  vector<string> list;
  vector<string> target;
  int iOut, errCode;
  string cell;
  string typeCode;
  int len;
  
#define NS 10
  regex_t preg;
  regmatch_t pmatch[NS];
  int icmp = 0;
  int errCodeRegex;
  string patternString;
  bool keyNotFound;
  vector<pair<unsigned int, unsigned int> > duplRows;
/***************************/


  /************** PARENT-CHILD **************/
  // For parent-chaild relationship checking we need table "item_linked"
  // index on "child_name" column, after search value from "parent_name" column
  ParentChildTbl = ref->GetTablePtr("item_linked");
  list.clear();
  list.push_back("child_name");
  int errCodeP_CH;
  // Creates an index on column child_name, table item_linked ddl/dictionary
  ParentChildTbl->CreateIndex("index0",list);
  ParentIndex = ParentChildTbl->GetColumnIndex("parent_name");


  /************** MANDATORY **************/
  // For mandatory checking we need tables "item" from ref
  // Table ref->Item - index on columns "catgory_id"+"mandatory_code",
  // after search value from "name" column
  // then checking for particular item in ctegory
  vector<string> nameList;

  ISTable * refItem;
  vector<string> refItemList;
  vector<string> refItemTarget;
  int refItemIndex;

  vector<string> inItemList;
  vector<string> inItemTarget;

  nameList.clear();
  nameList.push_back("name");

  refItem = ref->GetTablePtr("item");
  refItemList.clear();
  refItemList.push_back("category_id");
  refItemList.push_back("mandatory_code");
  refItem->CreateIndex("index0",refItemList);
  refItemIndex = refItem->GetColumnIndex("name");

  /************** REGEX **************/
  // For regular expresson checking we need table "item_type"
  // index on column "name", after search value from "code" column
  // and table "item_type_list" (to find reg. exp.)
  // index on column "code, after search value from "construct" column

  ISTable * ItemType,*ItemTypeL;
  vector<string> ItemTypeLList;
  vector<string> ItemTypeLTarget;
  int ItemTypeIndex, ItemTypeLIndex;

  ItemType = ref->GetTablePtr("item_type");
  ItemTypeIndex = ItemType->GetColumnIndex("code");

  ItemTypeL = ref->GetTablePtr("item_type_list");
  ItemTypeLList.clear();
  ItemTypeLList.push_back("code");
  ItemTypeLIndex = ItemTypeL->GetColumnIndex("construct");

  
  /************** RANGE **************/
  // For range checking we need table "item_range"
  // index on column "name", after search value from "maximum" & "minimum" columns
  ISTable * ItemRange;
  int ItemRangeIndex1, ItemRangeIndex2;
  int errCodeRange;
  string* maxlist = NULL;
  string* minlist = NULL;
  int irange = 0;
  int matched;
  double dval, dmin, dmax;
  int ival, imin, imax;

  ItemRange = ref->GetTablePtr("item_range");
  ItemRangeIndex1 = ItemRange->GetColumnIndex("maximum");
  ItemRangeIndex2 = ItemRange->GetColumnIndex("minimum");


  /************** ENUMERATION **************/
  // For enumeration checking we need table "item_enumeration"
  // index on column "name", after search value from "value" column
  ISTable * ItemEnumeration;
  int ItemEnumerationIndex;
  int errCodeEnumeration;
  string* enumlist = NULL;
  int ienum = 0;
  int ie;
  double de;

  ItemEnumeration = ref->GetTablePtr("item_enumeration");
  ItemEnumerationIndex = ItemEnumeration->GetColumnIndex("value");

  /************** REDUNDANT ROWS - KEY **************/
  // For renduntant rows checking we need table "category_key"
  // index on column "id", after search value from "name" column
  ISTable * CategoryKey;
  int KeyIndex;
  vector<string> keyTarget;
  vector<string> keyList;
  keyList.clear();
  keyList.push_back("id");

  CategoryKey = ref->GetTablePtr("category_key");
  KeyIndex = CategoryKey->GetColumnIndex("name");
  vector<string> colNamesKey;


  in->GetBlockNames(BlockNames);
  numBlocks = BlockNames.size();
  
  for (i=0; i<numBlocks; i++) { //do for all datablocks
    in->GetTableNames(TableNames, BlockNames[i]);
    numTables = TableNames.size();
    for (j=0; j<numTables; j++) { //do for all tables
      Category = in->GetTablePtr(BlockNames[i], TableNames[j]);
      /************** REDUNDANT ROWS - KEY **************/
      keyTarget.clear();
      keyTarget.push_back(TableNames[j]);
      CategoryKey->Search(keyTarget,keyList,OutList,errCode);
      if (errCode>=0) {
	colNamesKey.clear();
	for (k=0; k<(int)(OutList.size()); k++) {
	  CategoryKey->GetCell(cell,KeyIndex,OutList[k]);
	  CifFileObjBase::GetKeywordFromItemName(keyword,cell);
	  colNamesKey.push_back(keyword);
	}

        keyNotFound = false;
        for (unsigned int i = 0; i < colNamesKey.size(); i++)
        {
            if (Category->GetColumnIndex(colNamesKey[i]) ==
              COLUMN_NAME_NOT_FOUND)
            {
                log << "ERROR - key column not found in table " <<
                    Category->GetName() << " - " << colNamesKey[i] << endl;
                keyNotFound = true;
                break;
            }
        }

        if (!keyNotFound)
        {
            Category->FindRedundantRows(colNamesKey, duplRows, 1);
            if (!duplRows.empty())
            {
                for (unsigned int rowI = 0; rowI < duplRows.size(); rowI++)
                {
                    log << "ERROR - duplicate row in table "<<
                      Category->GetName() << ". Row " << duplRows[rowI].first
                      << " is duplicate of row "<< duplRows[rowI].second <<
                      endl;
                }
            }
        }
      }
      /**************************************************/
      /*************** MANDATORY **************/
      refItemTarget.clear();
      refItemTarget.push_back(TableNames[j]);
      refItemTarget.push_back("yes");
      refItem->Search(refItemTarget,refItemList,OutList,errCode);
      if (errCode>=0) {
	for (k=0; k<(int)(OutList.size()); k++) {
	  refItem->GetCell(cell,refItemIndex,OutList[k]);
	  CifFileObjBase::GetKeywordFromItemName(keyword,cell);
	  errCode = Category->GetColumnIndex(keyword);
	  if (errCode<0) {
	    log<<"ERROR - Mandatory item "<<cell<<" is not in datablock " << BlockNames[i] <<endl;
	    ret = 0;
	  }
	}
      }
      /****************************************/
      
      CifFileObjBase::GetItemNames(ItemNames, *Category);
      for (k=0; k<(int)(ItemNames.size()); k++) { //all items

	target.clear();
	target.push_back(ItemNames[k]);

	/************** PARENT-CHILD **************/

	// Checking if item has a parent
	iOut = ParentChildTbl->FindFirst(target,list,errCodeP_CH);
	if (errCodeP_CH>=0) {
	  // if YES checking for "parent" table & value in "child" item
	  ParentChildTbl->GetCell(ParentName,ParentIndex,iOut);
	  CifFileObjBase::GetCategoryNameFromItemName(ParentTblName,ParentName);
	  if (!in->IsTablePresent(BlockNames[i],ParentTblName)) {
	    //table for parent item does not exist
	    log <<"WARNING - Item "<<ItemNames[k]<<" has parent item "<< ParentName<<" but table "
		<< ParentTblName <<" is NOT present in the file"<<endl;
            // Set indication that the parent table does not exist. This will
            // prevent step 2 of parent-child relationship to be executed.
            errCodeP_CH = NOT_FOUND;
	  }
	  else {
            // Set indication that the parent table exists.
            errCodeP_CH = NO_TABLE_ERROR;
	    ParentTbl = in->GetTablePtr(BlockNames[i],ParentTblName);
	    CifFileObjBase::GetKeywordFromItemName(keyword,ParentName);
	    ParentList.clear();
	    ParentList.push_back(keyword);
	    if(!ParentTbl->IndexExists(keyword)) { 
	      ParentTbl->CreateIndex(keyword,ParentList);
	    }
	  }
	}

	/************** REGEX step 1**************/
	/* compile a regex */
	iOut = ItemType->FindFirst(target,nameList,errCodeRegex);
	if (errCodeRegex>=0) {
	  ItemType->GetCell(typeCode,ItemTypeIndex,iOut);
	  ItemTypeLTarget.clear();
	  ItemTypeLTarget.push_back(typeCode);
	  iOut = ItemTypeL->FindFirst(ItemTypeLTarget,ItemTypeLList,errCode);
	  ItemTypeL->GetCell(cell,ItemTypeLIndex,iOut);
          ConvertEscapedString(cell, patternString); 
	  icmp = regcomp(&preg,patternString.c_str(),REG_EXTENDED);
          regfree(&preg);
	}
	/****************************************/
	/************** RANGE step 1**************/
	/* finding min & max */
	ItemRange->Search(target,nameList,OutList,errCodeRange);
	if (errCodeRange>=0) {
          // VLAD: Why OutList->size() + 1 and not just OutList->size()?
	  maxlist = (string*) new string[OutList.size() + 1];
	  minlist = (string*) new string[OutList.size() + 1];
	  irange = OutList.size();
	  for (l=0;l<irange;l++) {
	    ItemRange->GetCell(cell,ItemRangeIndex1,OutList[l]);
	    maxlist[l] = cell;
	    ItemRange->GetCell(cell,ItemRangeIndex2,OutList[l]);
	    minlist[l] = cell;
	  }
	}
	/****************************************/
	/************ ENUMERATION step 1************/
	ItemEnumeration->Search(target,nameList,OutList,errCodeEnumeration);
	if (errCodeEnumeration>=0) {
	  enumlist = (string*) new string[OutList.size() + 1];
	  ienum = OutList.size();
	  for (l=0;l<ienum;l++) {
	    ItemEnumeration->GetCell(cell,ItemEnumerationIndex,OutList[l]);
	    enumlist[l] = cell;
	  }
	}
	/****************************************/
	// loop for all rows
	numRows = Category->GetNumRows();
	for (l=0; l<numRows ;l++){

	  Category->GetCell(cell,k,l);
	  /*************** PARENT-CHILD step 2**************/
	  if (errCodeP_CH==0) {
	    if ((cell!=".") && (cell!="?") && (cell!="")) {
	      ParentTarget.clear();
	      ParentTarget.push_back(cell);
	      ParentTbl->FindFirst(ParentTarget,ParentList,errCode);
	      if (errCode<0) {
		log<<"ERROR - Value "<< cell <<" for "<<ItemNames[k]
		   <<" is NOT defined in parent item "<< ParentName <<endl;
		ret = 0;
	      }
	    }
	  }
	  /****************************************/
	  
	  /*************** REGEX step 2**************/
	  if (errCodeRegex>=0) {
	    if (icmp==0 && (cell!=".") && (cell!="?") && (cell!="")) {
	      regcomp(&preg,patternString.c_str(),REG_EXTENDED);
	      ret = regexec(&preg,cell.c_str(),NS,pmatch,0);
              regfree(&preg);
	      if (ret!=0) {
		log<<"ERROR - value "<<cell<<" for "<<ItemNames[k]<<" does not match the pattern (row "<<l<<")"<<endl;
		ret = 0;
	      }
	      else { 
		len=pmatch[0].rm_eo - pmatch[0].rm_so;
		if (len!=(int)(cell.size())&& (typeCode != "text")) { 
		  log<<"ERROR - value "<<cell<<" for "<<ItemNames[k]<<" does not match the pattern (row "<<l<<")"<<endl;
		  log<<"ERROR - for "<<ItemNames[k]<<" only "<<cell.substr(pmatch[0].rm_so)<<" is matched (row "<<l <<")"<<endl;
		  ret = 0;
		}
	      }
	    }
	    /****************************************/
	    /************** RANGE step 2**************/
	    //	    Range(cell,maxlist,minlist,
	    if (errCodeRange>=0&& (cell!=".") && (cell!="?") && (cell!="")) {
	  m=0;
	  matched = 0;
	  if (typeCode == "float") {
	    dval = String::StringToDouble(cell);
	    while (m<irange && !matched) {
	      if (((minlist[m] != "?") && (minlist[m] != ".")
                && (!minlist[m].empty())) && ((maxlist[m] != "?")
                && (maxlist[m] != ".") && (!maxlist[m].empty()))) {
		dmin = String::StringToDouble(minlist[m]);
		dmax = String::StringToDouble(maxlist[m]);
		if (dmin==dmax) {
		  if (dval==dmax) matched = 1;
		}
		else
		  if (dval > dmin && dval < dmax) matched = 1;
	      }
	      else {
		if ((minlist[m] != "?") && 
                  (minlist[m] != ".") && 
                  (!minlist[m].empty())) {
		  dmin = String::StringToDouble(minlist[m]);
		  if (dval > dmin) matched = 1;
		}
		else {
		  if ((maxlist[m] != "?") && 
                    (maxlist[m] != ".") && 
                    (!maxlist[m].empty())) {
		    dmax = String::StringToDouble(maxlist[m]);
		    if (dval < dmax) matched = 1;
		  }
		  else
		    matched = 1;
		}
	      }
	      m++;
	    }
	    if (matched==0) {
	      log<<"ERROR - for "<<ItemNames[k]<<" "<<cell<<" is NOT in a range (row "<<l <<")"<<endl;
	      ret = 0;
	    }
	  }
	  else {
	    m=0;
	    matched = 0;
	    if (typeCode == "int") {
	      ival = String::StringToInt(cell);
	      while (m<irange && !matched) {
	      if (((minlist[m] != "?") && (minlist[m] != ".") 
                && (!minlist[m].empty())) && ((maxlist[m] != "?")
                && (maxlist[m] != ".") && (!maxlist[m].empty())))
              {
		imin = String::StringToInt(minlist[m]);
		imax = String::StringToInt(maxlist[m]);
		if (imin==imax) {
		  if (ival==imin) matched = 1;
		}
		else
		  if (ival > imin && ival < imax) matched = 1;
	      }
	      else {
		if ((minlist[m] != "?") && 
                  (minlist[m] != ".") && 
                  (!minlist[m].empty())) {
		  imin = String::StringToInt(minlist[m]);
		  if (ival > imin) matched = 1;
		}
		else {
		  if ((maxlist[m] != "?") && 
                    (maxlist[m] != ".") && 
                    (!maxlist[m].empty())){
		    imax = String::StringToInt(maxlist[m]);
		    if (ival < imax) matched = 1;
		  }
		  else
		    matched = 1;
		}
	      }
		m++;
	      }
	      if (matched==0) {
	      log<<"ERROR - for "<<ItemNames[k]<<" "<<cell<<" is NOT in a range (row "<<l <<")"<<endl;
	      ret = 0;
	      }
	    }
	  }
	}
	  /****************************************/
	  /************ ENUMERATION step 2************/
	  if (errCodeEnumeration>=0 && (cell!=".") && (cell!="?") && (cell!="")) {
	    m=0;
	    matched = 0;
	    if (typeCode == "float") {
	      dval = String::StringToDouble(cell);
	      while (m<ienum && !matched) {
		de = String::StringToDouble(enumlist[m]);
		if (dval == de) matched = 1;
		m++;
	      }
	    }
	    else {
	      if (typeCode == "int") {
		ival = String::StringToInt(cell);
		while (m<ienum && !matched) {
		  ie = String::StringToInt(enumlist[m]);
		  if (ival == ie) matched = 1;
		  m++;
		}
	      }
	      else {
		while (m<ienum && !matched) {
		if (cell == enumlist[m]) matched = 1;
		m++;
		}
	      }
	    }
	    if (matched==0) {
	      log<<"ERROR - for "<<ItemNames[k]<<" "<<cell<<" is NOT defined in the enumeration list (row "<<l <<")"<<endl;
	      ret = 0;
	    }
	  }
	  /****************************************/
	  }
	} //for all rows
	
	if (errCodeRange>=0) {
	  if (maxlist) delete [] maxlist;
	  if (minlist) delete [] minlist;
	}
	if (errCodeEnumeration>=0) {
	  if (enumlist) delete [] enumlist;
	}
      
      } //for all items
      
      ItemNames.clear();
    }
    TableNames.clear();
  }
  
  BlockNames.clear();

  return ret;
}



int CifFileObjBase::DataChecking(CifFileObjBase *ref) {
  int ret1, ret2;
  ofstream log;
  log.open("Diagnostics.log",ios::out|ios::app);
  ret1 = MandatoryCategoryChecking(this,ref,log);
  ret2 = ItemsChecking(this,ref,log);
  log.close();
  if (ret1==0) 
    return ret2;
  else
    return ret1;
}
