/*
FILE:     STable.C
*/
/*
VERSION:  7.001
*/
/*
DATE:     5/18/2006
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2006 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/


/* 
  PURPOSE:    Class for string table
*/


#include <string>
#include <iostream>
#include <iomanip>

#include "GenString.h"
#include "STable.h"


using namespace std;


STable::STable(const StringCompare::eCompareType colCaseSense)
{

    Init(String::EmptyString, colCaseSense);

}


STable::STable(const string& name,
  const StringCompare::eCompareType colCaseSense)
{

    Init(name, colCaseSense);

}


STable::~STable()
{

    Clear();

}


void STable::Init(const string& name,
  const StringCompare::eCompareType colCaseSense)
{

    _tableName = name;
    _numRows = 0;
    _colCaseSense = colCaseSense;

}


void STable::Copy(const STable& inTable)
{

    _tableName = inTable._tableName;
    _numRows = inTable._numRows;
    _colNames = inTable._colNames;
    _colCaseSense = inTable._colCaseSense;
    // VLAD: VERIFY: WHAT ARE THE CONSEQUENCES OF THE BELOW STATEMENT
    // HOW ARE CONTAINERS OF POINTERS HANDLED
    _data = inTable._data;

    for (unsigned int colIndex = 0; colIndex < _colNames.size(); colIndex++)
    {
        _data[colIndex] = new vector<string>(*(inTable._data[colIndex]));
    }

}


void STable::Rename(const string& name)
{

    _tableName = name;

}


int STable::AddColumn(const string& colName)
{

    return(InsertColumn(colName, _colNames.size()));

}


int STable::InsertColumn(const string& colName, const unsigned int colIndex)
{

    if (colName.empty())
    {
        // Empty column name. Return error.
        return(-1);
    }

    if (colIndex > _colNames.size())
    {
        // Wrong column index. Return error.
        return(-1);
    }

    int newColIndex = GetColumnIndex(colName);
    if (newColIndex >= 0)
    {
        // Found a column with the same name. Return error.
        return(-1);
    }

    _data.insert(_data.begin() + colIndex, new vector<string>());

    _colNames.insert(_colNames.begin() + colIndex, colName);

    return(_colNames.size());

}


int STable::InsertColumn(const vector<string>& col, const string& colName,
  const unsigned int colIndex)
{

    // VLAD: ERROR HANDLING CODE - ADD
    InsertColumn(colName, colIndex);

    FillColumn(col, colIndex);

    return(colIndex);

}


int STable::FillColumn(const vector<string>& col,
  const unsigned int colIndex)
{

    if (col.empty())
        return ADD_UPDATE_NULL;

    // out of range
    if (colIndex >= _colNames.size())
        return COLUMN_OUT_OF_BOUNDS;

    if (_data[colIndex] != NULL)
    {
        if (col.size() < _data[colIndex]->size())
        {
            // VLAD: CHECK IF THIS IS SAFE. JUST REPLACE THE BOTTOM WITH
            // CLEAR AND ADD COMMENTS HERE TO UNDERSTAND WHAT IS GOING ON
            _data[colIndex]->erase(_data[colIndex]->begin(),
              _data[colIndex]->begin() + col.size());
            copy(col.begin(), col.end(), _data[colIndex]->begin());
        }
        else
        {
            _data[colIndex]->clear();
            copy(col.begin(), col.end(), back_inserter(*_data[colIndex]));
        }
    }
    else
    {
        _data[colIndex] = new vector<string>();
        *_data[colIndex] = col;
    }

    if (_data[colIndex]->size() > _numRows)
    {
        _numRows = _data[colIndex]->size();
    }

    return NO_TABLE_ERROR;

}


int STable::AppendToColumn(const vector<string>& col,
  const string& colName)
{

    int colIndex = GetColumnIndex(colName);

    return AppendToColumn(col, colIndex);

}


int STable::AppendToColumn(const vector<string>& col,
  const unsigned int colIndex)
{

    if (col.empty()) return ADD_UPDATE_NULL;

    if (colIndex >= _colNames.size())
       return COLUMN_OUT_OF_BOUNDS;

    // VLAD: EXAMINE: Is this safe?
    copy(col.begin(), col.end(), _data[colIndex]->end());

    if (_data[colIndex]->size() > _numRows)
      _numRows = _data[colIndex]->size();

    return NO_TABLE_ERROR;

}


int STable::AddCellToColumn(const string& cell, const unsigned int colIndex)
{

    if (cell.empty()) return ADD_UPDATE_NULL;

    if (colIndex >= _colNames.size())
        return COLUMN_OUT_OF_BOUNDS;

    _data[colIndex]->push_back(cell);

    if (_data[colIndex]->size() > _numRows)
        _numRows = _data[colIndex]->size();

    return NO_TABLE_ERROR;

}


int STable::ClearColumn(const unsigned int colIndex)
{

    if (colIndex >= _colNames.size())
        return COLUMN_OUT_OF_BOUNDS;

    bool reCalcNumRows = false;
    if (_data[colIndex]->size() == _numRows)
    {
        // The column that is about to be cleared has the most elements
        // and its size is the number of rows. Set a flag that the number
        // of rows has to be re-calculated upon column clearing.
        reCalcNumRows = true;
    }

    // VLAD: EXAMINE: Can this be done without for loop?
    for (unsigned int rowIndex = 0; rowIndex < _data[colIndex]->size();
      rowIndex++)
    {
        (*_data[colIndex])[rowIndex].clear();
    }

    _data[colIndex]->clear();

    if (reCalcNumRows)
    {
        // Determines the maximum size of all columns and sets it to
        // _numRows
        unsigned int maxColSize = 0;
        for (unsigned int colIndex = 0; colIndex < _colNames.size();
          colIndex++)
        {
            if (_data[colIndex]->size() > maxColSize)
            {
                maxColSize = _data[colIndex]->size();
            }
        }

        _numRows = maxColSize;
    }

    return NO_TABLE_ERROR;

}


int STable::DeleteColumn(const unsigned int colIndex)
{

    // VLAD: EXAMINE: Does this have to update _numRows? No, since
    // ClearColumn() is doing it. Check also other column base and in
    // general all other methods for _numRows updating.
    if (colIndex >= _colNames.size())
        return COLUMN_OUT_OF_BOUNDS;

    ClearColumn(colIndex);

    delete _data[colIndex];

    _data.erase(_data.begin() + colIndex);

    _colNames.erase(_colNames.begin() + colIndex);

    return NO_TABLE_ERROR;

}
 

int STable::RenameColumn(const string& oldColName, const string& newColName)
{

    if (oldColName.empty() || newColName.empty())
    {
        // If any of the column names is empty, return error.
        return(-1);
    }

    int colIndex = GetColumnIndex(newColName);
    if (colIndex >= 0)
    {
        // Found a column with the same name as new column. Return error.
        return(-1);
    }

    colIndex = GetColumnIndex(oldColName);
    if (colIndex < 0)
    {
        // Old column not found. Return error.
        return(-1);
    }

    _colNames[colIndex] = newColName;

    return(colIndex);

}


void STable::GetColumnNames(vector<string>& colNames)
{

    colNames = _colNames;

}


void STable::GetColumn(vector<string>& col, const unsigned int colIndex)
{

    // out of range
    if (colIndex >= _colNames.size())
    {
        return;
    }

    if (_data[colIndex] != NULL)
    {
        col = *_data[colIndex];
    }

}


void STable::GetSubColumn(vector<string>& subCol,
  const unsigned int colIndex, const unsigned int fromRowIndex,
  unsigned int toRowIndex)
{

    if (colIndex >= _colNames.size())
        return;

    if (toRowIndex < fromRowIndex)
        return;

    if (fromRowIndex >= _data[colIndex]->size())
        return;

    if (toRowIndex >= _data[colIndex]->size())
        toRowIndex = _data[colIndex]->size() - 1;

    // toRowIndex is inclusive, so one more spot for it needs to be reserved
    subCol.reserve(toRowIndex - fromRowIndex + 1);

    copy((*_data[colIndex]).begin() + fromRowIndex,
      (*_data[colIndex]).begin() + toRowIndex + 1,
      back_inserter(subCol));

}


void STable::GetSubColumn(vector<string>& subCol,
  const unsigned int colIndex, const vector<unsigned int>& rowIndex)
{

    if (colIndex >= _colNames.size()) return;

    subCol.reserve(rowIndex.size());

    for (unsigned int index = 0; index <  rowIndex.size(); index++)
    {
        // VLAD: SHOULDNT THIS BE AN ERROR
        if (rowIndex[index] >= _data[colIndex]->size())
            continue;
        subCol.push_back((*_data[colIndex])[rowIndex[index]]);
    }

}


int STable::GetColumnIndex(const string& colName)
{

    unsigned int colIndex;

    // VLAD: GENERALIZE THIS TO INVOKE SOME SEARCH FUNCTION AND LATER
    // YOU CAN CHANGE SEARCH ALGORITHM AS YOU WISH
    // VLAD: XSTRING: Replace all this code with some elegant code
    for (colIndex = 0; colIndex < _colNames.size(); colIndex++)
    {
        if (String::IsEqual(_colNames[colIndex], colName, _colCaseSense))
        {
            break;
        }
    }

    if (colIndex == _colNames.size())
    {
        return(COLUMN_NAME_NOT_FOUND);
    }
    else
    {
        return(colIndex);
    }

}


int STable::GetColumnsIndices(const vector<string>& colNames, 
  vector<unsigned int>& colIndices)
{

    int colIndex;

    for (unsigned int index = 0; index < colNames.size(); index++)
    {
        colIndex = GetColumnIndex(colNames[index]);
        if (colIndex >= 0)
            colIndices.push_back(colIndex);
        else
            return(SOME_COLUMN_NAMES_NOT_FOUND);
    }

    return(NO_TABLE_ERROR);

}


int STable::AddRow()
{

    return(InsertRow(_numRows));

}


int STable::InsertRow(const unsigned int rowIndex)
{

    if (rowIndex > _numRows) return ROW_OUT_OF_BOUNDS;

    vector<string>::iterator pos;

    unsigned int numInsCells = 1;

    for (unsigned int colIndex = 0; colIndex < _colNames.size(); colIndex++)
    {
        if (rowIndex < _data[colIndex]->size())
        {
            pos = _data[colIndex]->begin() + rowIndex;
            numInsCells = 1;
        }
        else
        {
            pos = _data[colIndex]->end();
            numInsCells = rowIndex + 1 - _data[colIndex]->size();
        }

        _data[colIndex]->insert(pos, numInsCells, String::EmptyString);
    }

    _numRows++;

    return(_numRows);

}


int STable::InsertRow(const vector<string>& row,
  const unsigned int rowIndex)
{

    int ret = InsertRow(rowIndex);

    if (ret < TABLE_WARNING) return ret;

    ret = FillRow(row, rowIndex);

    if (ret < TABLE_WARNING) return ret;

    return(_numRows);

}


int STable::FillRow(const vector<string>& row,
  const unsigned int rowIndex)
{

    if (_numRows < rowIndex) return ROW_OUT_OF_BOUNDS;

    unsigned int len = row.size();

    if (len > _colNames.size())
        len = _colNames.size();

    for (unsigned int colIndex = 0; colIndex < len; colIndex++)
    {
        if (_data[colIndex]->size() <= rowIndex)
        {
            _data[colIndex]->insert(_data[colIndex]->end(),
              rowIndex + 1 - _data[colIndex]->size(), String::EmptyString);
        }
        (*_data[colIndex])[rowIndex] = row[colIndex];
    }

    return NO_TABLE_ERROR;

}


int STable::AppendToRow(const vector<string>& row,
  const unsigned int rowIndex)
{

    // VLAD: ERROR HANDLING
    if (rowIndex > _numRows)
        return ROW_OUT_OF_BOUNDS;
    if (row.empty())
        return ADD_UPDATE_NULL;

    unsigned int fromColIndex, toColIndex;

    GetRowAppendRange(row, rowIndex, fromColIndex, toColIndex);

    FillRow(row, rowIndex, fromColIndex, toColIndex);

    return NO_TABLE_ERROR;

}


int STable::AddCellToRow(const string& cell, const unsigned int rowIndex)
{

    if (rowIndex > _numRows)
        return ROW_OUT_OF_BOUNDS;

    unsigned int colIndex;

    GetCellAddIndex(rowIndex, colIndex);

    (*_data[colIndex])[rowIndex] = cell;

    return NO_TABLE_ERROR;

}


int STable::ClearRow(const unsigned int rowIndex)
{

    if (rowIndex >= _numRows) return ROW_OUT_OF_BOUNDS;

    for (unsigned int colIndex = 0; colIndex < _colNames.size(); colIndex++)
    {
        if (rowIndex < _data[colIndex]->size())
        {
            (*_data[colIndex])[rowIndex].clear();
        }
    }

    return NO_TABLE_ERROR;

}


int STable::DeleteRow(const unsigned int rowIndex)
{

    if (rowIndex >= _numRows) return ROW_OUT_OF_BOUNDS;

    for (unsigned int colIndex = 0; colIndex < _colNames.size(); colIndex++)
    {
        if (rowIndex < _data[colIndex]->size())
        {
            _data[colIndex]->erase(_data[colIndex]->begin() + rowIndex);
        }
    }

    return _numRows--;

}


void STable::GetRow(vector<string>& row, const unsigned int rowIndex)
{

    GetSubRow(row, rowIndex, 0, _colNames.size() - 1);

}


void STable::GetSubRow(vector<string>& subRow, const unsigned int rowIndex,
  const unsigned int fromColIndex, unsigned int toColIndex)
{

    if (rowIndex >= _numRows) return;
    if (toColIndex < fromColIndex) return;
    if (fromColIndex >= _colNames.size()) return;
    if (toColIndex >= _colNames.size()) toColIndex = _colNames.size() - 1;

    // toColIndex is inclusive, so one more spot for it needs to be reserved
    subRow.reserve(toColIndex - fromColIndex + 1);

    for (unsigned int colIndex = fromColIndex; colIndex <= toColIndex;
      colIndex++)
    {
        if (rowIndex < _data[colIndex]->size())
        {
            subRow.push_back((*_data[colIndex])[rowIndex]);
        }
        else
        {
            subRow.push_back(String::EmptyString);
        }
    }

}


int STable::UpdateCell(const string& cell, const unsigned int colIndex,
  const unsigned int rowIndex)
{

    if (colIndex >= _colNames.size())
        return COLUMN_OUT_OF_BOUNDS;

    if (rowIndex > _numRows) return ROW_OUT_OF_BOUNDS;

    // VLAD: DESIGN: THINK ABOUT IT. THE NAME UPDATE SUGGEST THAT SOMETHING
    // EXISTS. SIDEEFFECT OF THIS FUNCTION IS THAT NUMBER OF ROWS CAN BE
    // INCREASED. THAT IS BAD. AND THIS BEHAVIOUR IS DIFFERENT THAN ISTABLE
    // THIS ADD ROW SHOULD NOT BE ALLOWED
    if (rowIndex == _numRows) AddRow();

    // WHY DO WE ALLOW THIS. SHOULDNT THIS BE AN ERROR
    if (rowIndex >= _data[colIndex]->size())
    {
        // VLAD: FIX THIS WITH AN INSERT OF MULTIPLE EMPTY STRINGS AND
        // REMOVE THE FOR LOOP
        for (unsigned int i = _data[colIndex]->size(); i <= rowIndex; i++)
        {
            _data[colIndex]->push_back(String::EmptyString);
        }
    }
 
    (*_data[colIndex])[rowIndex] = cell;

    return NO_TABLE_ERROR;

}


int STable::GetCell(string& cell, const unsigned int colIndex,
  const unsigned int rowIndex)
{

    // VLAD: SIMPLIFICATION: CREATE A FUNCTION TO CHECK BOUNDS OF
    // COLUMN INDICES AND ROW INDICES
    if (colIndex >= _colNames.size())
        return COLUMN_OUT_OF_BOUNDS;
    if (rowIndex >= _data[colIndex]->size())
        return ROW_OUT_OF_BOUNDS;

    cell = (*_data[colIndex])[rowIndex];

    return NO_TABLE_ERROR;

}


void STable::PrintTable()
{

    unsigned int i, j;


    cout << endl << "Table " << _tableName <<  "  Columns: " <<
      _colNames.size() << "  Rows: " << _numRows << endl << endl;

    for (i = 0; i < _colNames.size(); i++)
    {
        cout << setw(10) << _colNames[i] << " ";
    }
    cout << endl;

    for (j = 0; j < _numRows; j++)
    {
        for (i = 0; i < _colNames.size(); i++)
        {
            if (_data[i]->size() > j)
            {
                cout << setw(10) << (*(_data[i]))[j] << " ";
            }
            else
            {
                cout << setw(10) << " " << " ";
            }
        }
    
        cout << endl;
    }

}


void STable::Clear()
{

    _tableName.clear();

    _colNames.clear();

    if (!_data.empty())
    {
        for (unsigned int colIndex = 0; colIndex < _data.size();
          colIndex++)
        {
            if (_data[colIndex] != NULL)
	    {
	        _data[colIndex]->clear();
	        delete _data[colIndex];
	    }
        }

        _data.clear(); 
    }

    _numRows = 0;

    _colCaseSense = StringCompare::eCASE_SENSITIVE;

}


int STable::VerifyColumnsIndices(const vector<unsigned int>& colIndices)
{

    for (unsigned int index = 0; index < colIndices.size(); index++)
    {
        if (colIndices[index] >= _colNames.size())
        {
            return(COLUMN_OUT_OF_BOUNDS);
        }
    }

    return(NO_TABLE_ERROR);

}


void STable::GetRowAppendRange(const vector<string>& row,
  const unsigned int rowIndex, unsigned int& fromColIndex,
  unsigned int& toColIndex)
{

    // VLAD: INTRODUCE PAIRS AND RETURN from and to in STL PAIR
    // VLAD: ERROR HANDLING return ROW_OUT_OF_BOUNDS
    if (_numRows < rowIndex) return;

    // VLAD:BUG. What should be the initial value of i? It must be
    // initialized to something, since it can be used uninitialized.
    unsigned int i = 0;

    for (unsigned int j = 0; j < _colNames.size(); j++)
    {
        i = _colNames.size() - 1 - j;

        if (rowIndex < _data[i]->size())
        {
            if (!((*_data[i])[rowIndex].empty()))
            {
                // VLAD: BUG if i == _colNames.size()
                i++;
                break;
            }   
        }
        else
        {
            _data[i]->insert(_data[i]->end(), rowIndex + 1 - _data[i]->size(),
              String::EmptyString);
        }
    }

    fromColIndex = i;

    toColIndex = fromColIndex + row.size() - 1;

    if (toColIndex >= _colNames.size())
        toColIndex = _colNames.size() - 1;

}


int STable::FillRow(const vector<string>& row,
  const unsigned int rowIndex, const unsigned int fromColIndex,
  const unsigned int toColIndex)
{

    // VLAD ERROR HANDLING FIX ALL THIS
    if (_numRows < rowIndex) return ROW_OUT_OF_BOUNDS;
    if (toColIndex < fromColIndex) return ROW_OUT_OF_BOUNDS;
    if (fromColIndex >= _colNames.size()) return ROW_OUT_OF_BOUNDS;
    if (toColIndex >= _colNames.size()) return ROW_OUT_OF_BOUNDS;

    unsigned int numCols;

    if ((toColIndex - fromColIndex + 1) > row.size())
    {
        numCols = row.size();
    }
    else
    {
        numCols = toColIndex - fromColIndex + 1;
    }

    for (unsigned int colIndex = fromColIndex;
      colIndex < (fromColIndex + numCols); colIndex++)
    {
        if (_data[colIndex]->size() <= rowIndex)
        {
            _data[colIndex]->insert(_data[colIndex]->end(),
              rowIndex + 1 - _data[colIndex]->size(),
              String::EmptyString);
        }
        (*_data[colIndex])[rowIndex] = row[colIndex - fromColIndex];
    }

    return NO_TABLE_ERROR;

}


int STable::GetCellAddIndex(const unsigned int rowIndex,
  unsigned int& colIndex)
{

    if (rowIndex > _numRows) return ROW_OUT_OF_BOUNDS;

    for (colIndex = 0; colIndex < _colNames.size(); colIndex++)
    {
        if (rowIndex < _data[colIndex]->size())
        {
            if ((*_data[colIndex])[rowIndex].empty())
            {
                break;
            }   
        }
        else
        {
            break;
        }
    }

    if (colIndex == _colNames.size()) return ROW_OUT_OF_BOUNDS; // Full row

    if (rowIndex >= _data[colIndex]->size())
    {
        _data[colIndex]->insert(_data[colIndex]->end(),
          rowIndex + 1 - _data[colIndex]->size(), String::EmptyString);
    }

    return NO_TABLE_ERROR;
}

