/*
FILE:     CifInterface.C
*/
/*
VERSION:  7.001
*/
/*
DATE:     5/18/2006
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2006 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
#include <string>
#include <stdlib.h>
#include "GenString.h"
#include "CifInterface.h"
#include "FileNavigatorError.h"
#include "DDLParserBase.h"
#include "DICParserBase.h"
#include "DICFileObj.h"
#include "DDLFileObj.h"
#include "regex.h"

int CifInterface::Init(const string& fileName, int inVerify)
/* ----------------------------------------------------------------------------
   Init(): Must be called before any other CifInterface method. Reads DDL
	which is used for dictionary validation
   --------------------------------------------------------------------------*/
{
  string diags;
  _ddlformat = new ISTable("ddlformat");
  if (fileName.empty()) return FILENAME_INVALID;
  filename = fileName;
  verify = inVerify;
  _ddl = new DDLFileObj(string("ddl"), WRITE_MODE, 80, "?", 1);
  if (!_ddl) return ALLOCATION_ERROR;
  else
  {
    DDLParser* ddlParserR = NULL;

    ddlParserR = new DDLParser(_ddl, _ddlformat, _ddl->GetVerbose());

    ddlParserR->Parse(filename, diags);

    delete(ddlParserR);

    _ddl->DataChecking(_ddl);
    return(0);
  }
}

int CifInterface::ReadFile(const string& fileName)
/* ----------------------------------------------------------------------------
   ReadFile(): Reads dictionary.
   --------------------------------------------------------------------------*/
{
  string diags;
  _dicformat = new ISTable("dicformat");
  _dic = new DICFileObj(string("dictionary.odb"), WRITE_MODE, 80, "?", 0);

  DICParser* dicParserR = NULL;

  dicParserR = new DICParser(_dic, _ddl, _dicformat, _dic->GetVerbose());

  dicParserR->Parse(fileName, diags);

  delete(dicParserR);

  _dic->Compress(_ddl);

  _dic->DataChecking(_ddl);

//  CifFiles->PrintDiagnostics();

  return (0);

}

int CifInterface::CountDataBlocks()
/* ----------------------------------------------------------------------------
   CountDataBlocks(): Counts the number of datablocks that were read in with
   Init() and ReadFile()
   --------------------------------------------------------------------------*/
{
  return _dic->GetNumBlocks();
}

void CifInterface::GetDataBlockNames(vector<string>& dataBlockNames)
/* ----------------------------------------------------------------------------
   GetDataBlockNames(): Returns a copy of the datablock names and also returns
   the number of datablocks in the reference numDataBlock
   --------------------------------------------------------------------------*/
{
  dataBlockNames.clear();
  _dic->GetBlockNames(dataBlockNames);
}


string CifInterface::GetDataBlockDictionaryName(const string& dataBlockName)
/* ----------------------------------------------------------------------------
   GetDataBlockDictionaryName(): returns the name of the dictionary used to 
   validate the datablock named by dataBlockName
	(DataBlockName for ddl)
   --------------------------------------------------------------------------*/
{
  return _ddl->GetFirstBlockName();
}


ISTable * CifInterface::GetTablePtr(const string& blockName, const string& tableName)
{
  return _dic->GetTablePtr(blockName.c_str(),tableName.c_str());
}




string CifInterface::GetItemName(const string& categoryName,
                                          const string& itemKeyword) 
/* ----------------------------------------------------------------------------
   GetItemName(): will take categoryName and itemKeyword and return a complete
   item name. Note that this is a purely mechanical method that does not check
   for the existence of the category or the keyword
   --------------------------------------------------------------------------*/
{
  string target;

  target += CIF_DATA_PREFIX_CHARACTER;
  target += categoryName;
  target += CIF_DATA_JOIN_CHARACTER;
  target += itemKeyword;

  return target;
}

string CifInterface::GetItemKeywordByName(const string& itemName)
/* ----------------------------------------------------------------------------
   GetItemKeywordByName() takes the complete item name itemName and returns its
   keyword portion. Note that this method is purely mechanical and does not
   check for the existence of the item.
   --------------------------------------------------------------------------*/
{
  unsigned int i;
  string target(itemName);
  unsigned int length = target.size();
  for (i = 0; i < length; i++) {
    if (target[i] == CIF_DATA_JOIN_CHARACTER)
      break;
  }
  if (i == length) return "";
  else {
    // VLAD: POTENTIAL
    string ret(target, i + 1);

    return ret;
  }

}

string CifInterface::GetCategoryByItemName(const string& itemName)
/* ----------------------------------------------------------------------------
   GetCategoryByItemName() takes the complete item name itemName and returns
   the category portion. Note that this a purely mechanical method that does
   not verify the existence of the item.
   --------------------------------------------------------------------------*/
{
  unsigned int i;
  string target(itemName);
  unsigned int length = target.size();
  for (i = 0; i < length; i++) {
    if (target[i] == CIF_DATA_JOIN_CHARACTER)
      break;
  }
  if (i == length) return "";
  else {
    string temp = target;
    temp.erase(i, length - i);
    string ret(temp, 1);
    return ret;
  }

}

string CifInterface::GetItemDataTypeCode(const string& dataBlockName,
                                         const string& categoryName,
			                 const string& itemKeyword)
/* ----------------------------------------------------------------------------
   GetItemDataTypeCode(): returns the data type code of the item pointed to by
   dataBlockName, categoryName, and itemKeyword 
   --------------------------------------------------------------------------*/
{
  vector<string> ColumnNames;
  vector<string> values;
  vector<unsigned int> listOut;
  string cell;
  int errCode;
  int columnIndex;
  string ret;

  ISTable *tbl = _dic->GetTablePtr(dataBlockName, CIF_DDL_CATEGORY_ITEM_TYPE);
  if(tbl == NULL) return ret;
  ColumnNames.push_back(CIF_DDL_ITEM_NAME);
  string name = GetItemName(categoryName, itemKeyword);
  values.push_back(name);    

  columnIndex = tbl->GetColumnIndex(CIF_DDL_ITEM_NAME);
  tbl->SetFlags(ISTable::DT_STRING | ISTable::CASE_INSENSE,columnIndex);

  tbl->Search(values,ColumnNames,listOut,errCode);
  if (errCode<0) return ret;
  else {
	 columnIndex = tbl->GetColumnIndex(CIF_DDL_ITEM_CODE);
	 tbl->GetCell(cell,columnIndex,listOut[0]);
	 return cell;
  }
}

string CifInterface::GetItemDataTypeCodeInherit(const string& dataBlockName,
                                         const string& categoryName,
			                 const string& itemKeyword,
                                         const Family decendents, 
                                         int decendentsNum)
/* ----------------------------------------------------------------------------
   GetItemDataTypeCodeInherit(): returns the data type code of the item
   pointed to by dataBlockName, categoryName, and itemKeyword. If one is not
   found, the Family decendents will be searched so that an inherited data type
   code may be found. The number of decendents must also be supplied with
   decendentsNum.
   --------------------------------------------------------------------------*/
{
  string ret;
  int i;
  string category;
  string keyword;

  if (dataBlockName.empty() || categoryName.empty() || itemKeyword.empty()) return ret;
  if ((decendentsNum < 0) || ((decendentsNum > 0) && (!decendents)))
    decendentsNum = 0;

  ret = GetItemDataTypeCode(dataBlockName, categoryName, itemKeyword); 
  if (!ret.empty()) return ret;
  else {
    for (i = 0; i < decendentsNum; i++) {
      category = GetCategoryByItemName(decendents[i].name);
      keyword  = GetItemKeywordByName(decendents[i].name);
      ret = GetItemDataTypeCode(dataBlockName, category, keyword);
      if (!ret.empty()) break;
    }
  return ret;
  }
}


string CifInterface::GetItemDescription(const string& dataBlockName,
                                         const string& categoryName,
			                 const string& itemKeyword)
/* ----------------------------------------------------------------------------
   GetItemDescription(): returns the description of the item pointed to by 
   dataBlockName, categoryName, and itemKeyword, if it exists.
   --------------------------------------------------------------------------*/
{
  vector<string> ColumnNames;
  vector<string> values;
  vector<unsigned int> listOut;
  string ret;
  int columnIndex;
  int errCode;
  string cell;

  ISTable *tbl = _dic->GetTablePtr(dataBlockName, CIF_DDL_CATEGORY_ITEM_DESCRIPTION);
  if(tbl == NULL) return ret;
  ColumnNames.push_back(CIF_DDL_ITEM_NAME);
  string name = GetItemName(categoryName, itemKeyword);
  values.push_back(name);    

  columnIndex = tbl->GetColumnIndex(CIF_DDL_ITEM_NAME);
  tbl->SetFlags(ISTable::DT_STRING | ISTable::CASE_INSENSE,columnIndex);

  tbl->Search(values, ColumnNames, listOut, errCode);
  if (errCode<0) return ret;
  else {
	 columnIndex = tbl->GetColumnIndex(CIF_DDL_ITEM_DESCRIPTION);
	 tbl->GetCell(cell,columnIndex,listOut[0]);
	 return cell;
  }
}

string CifInterface::GetItemDescriptionNDB(const string& dataBlockName,
					   const string& categoryName,
					   const string& itemKeyword)
/* ----------------------------------------------------------------------------
   GetItemDescription(): returns the description of the item pointed to by 
   dataBlockName, categoryName, and itemKeyword, if it exists.
   --------------------------------------------------------------------------*/
{
  vector<string> ColumnNames;
  vector<string> values;
  vector<unsigned int> listOut;
  string ret;
  int columnIndex;
  int errCode;
  string cell;

  ISTable *tbl = _dic->GetTablePtr(dataBlockName, CIF_DDL_CATEGORY_NDB_ITEM_DESCRIPTION);
  if(tbl == NULL) return ret;
  ColumnNames.push_back(CIF_DDL_ITEM_NAME);
  string name = GetItemName(categoryName, itemKeyword);
  values.push_back(name);    

  columnIndex = tbl->GetColumnIndex(CIF_DDL_ITEM_NAME);
  tbl->SetFlags(ISTable::DT_STRING | ISTable::CASE_INSENSE,columnIndex);

  tbl->Search(values,ColumnNames,listOut,errCode);
  if (errCode<0) return ret;
  else {
	 columnIndex = tbl->GetColumnIndex(CIF_DDL_ITEM_DESCRIPTION);
	 tbl->GetCell(cell,columnIndex,listOut[0]);
	 return cell;
  }
}

string CifInterface::GetItemDescriptionInherit(const string& dataBlockName,
                                         const string& categoryName,
			                 const string& itemKeyword,
                                         const Family decendents, 
                                         int decendentsNum)
/* -------------------------------------------------------------------------
   GetItemDescriptionInherit(): returns the description of the item pointed to
   by dataBlockName, categoryName, and itemKeyword. If it does not exist, the
   Family decendents will be searched until an inheritable description is
   found. The number of decendents must also be specified by decendentsNum. 
   ------------------------------------------------------------------------- */
{
  string ret;
  int i;
  string category;
  string keyword;

  if (dataBlockName.empty() || categoryName.empty() || itemKeyword.empty()) return NULL;
  if (decendentsNum < 0 || (decendentsNum > 0 && decendents))
    decendentsNum = 0;

  ret = GetItemDescription(dataBlockName, categoryName, itemKeyword);
  if (!ret.empty()) return ret;
  else {
    for (i = 0; i < decendentsNum; i++) {
      category = GetCategoryByItemName(decendents[i].name);
      keyword  = GetItemKeywordByName(decendents[i].name);
      ret = GetItemDescription(dataBlockName, category, keyword);
      if (!ret.empty()) break;
  }
    return ret;
  }
}

string CifInterface::GetItemDescriptionInheritNDB(const string& dataBlockName,
						  const string& categoryName,
						  const string& itemKeyword,
						  const Family decendents, 
						  int decendentsNum)
/* -------------------------------------------------------------------------
   GetItemDescriptionInherit(): returns the description of the item pointed to
   by dataBlockName, categoryName, and itemKeyword. If it does not exist, the
   Family decendents will be searched until an inheritable description is
   found. The number of decendents must also be specified by decendentsNum. 
   ------------------------------------------------------------------------- */
{
  string ret;
  int i;
  string category;
  string keyword;

  if (dataBlockName.empty() || categoryName.empty() || itemKeyword.empty()) return ret;
  if (decendentsNum < 0 || (decendentsNum > 0 && decendents))
    decendentsNum = 0;

  ret = GetItemDescriptionNDB(dataBlockName, categoryName, itemKeyword); 
  if (!ret.empty()) return ret;
  else {
    for (i = 0; i < decendentsNum; i++) {
      category = GetCategoryByItemName(decendents[i].name);
      keyword  = GetItemKeywordByName(decendents[i].name);
      ret = GetItemDescriptionNDB(dataBlockName, category, keyword);
      if (!ret.empty()) break;
  }
    return ret;
  }
}

Family CifInterface::GetItemDecendency(const string& dataBlockName,
                                         const string& categoryName,
			                 const string& itemKeyword, int & numD)
/* -------------------------------------------------------------------------
   GetItemDecendency(): creates and returns a complete Family for the item
   pointed to by dataBlockName, categoryName, and itemKeyword.  The number of
   decendents in the Family is returned in the reference numD.
   ------------------------------------------------------------------------- */
{
  vector<string> ColumnNamesCh;
  vector<string> ColumnNamesPr;
  vector<string> values;
  vector<unsigned int> listOut;
  int columnIndex;
  int errCode;
  string cell;
  int num,len, num1;

  int * parentList, i;
  int pCount,totalCount, generationNum;
  numD = 0;
  pCount = 0;
  totalCount = 0;
  generationNum = 1;
  if (dataBlockName.empty() || categoryName.empty() || itemKeyword.empty()) return NULL;
  
  ISTable *tbl = _dic->GetTablePtr(dataBlockName, CIF_DDL_CATEGORY_ITEM_LINKED);
  if(tbl == NULL) return NULL;
  ColumnNamesCh.push_back(CIF_DDL_ITEM_CHILD_NAME);
  ColumnNamesPr.push_back(CIF_DDL_ITEM_PARENT_NAME);
  string name = GetItemName(categoryName, itemKeyword);
  values.push_back(name);   

  columnIndex = tbl->GetColumnIndex(CIF_DDL_ITEM_CHILD_NAME);
  tbl->SetFlags(ISTable::DT_STRING | ISTable::CASE_INSENSE,columnIndex);

  tbl->Search(values,ColumnNamesCh,listOut,errCode);
  if (errCode < 0) return NULL;
  
  columnIndex = tbl->GetColumnIndex(CIF_DDL_ITEM_PARENT_NAME);
  len = 2*tbl->GetNumRows()*sizeof(int)+1;
  parentList = (int *) malloc(len);

  num1 = listOut.size();
  for (i=0;i<num1; i++) {
	 parentList[2*pCount] = listOut[i];
	 parentList[2*pCount + 1] = generationNum;
	 pCount++;
  }
  while (totalCount<pCount) {
	 values.clear();
	 cell.clear();
	 tbl->GetCell(cell,columnIndex,parentList[2*totalCount]);
	 values.push_back(cell);
	 cell.clear();
	 tbl->Search(values,ColumnNamesCh,listOut,errCode);
	 if (errCode >=0) {
		generationNum = parentList[2*totalCount + 1] + 1;
		num = listOut.size();
		for (i=0;i<num; i++) {
		  parentList[2*pCount] = listOut[i];
		  parentList[2*pCount + 1] = generationNum;
		  pCount++;
		}
	 }
	 totalCount++;
  }

  Family f = (Family) malloc(totalCount*sizeof(struct family_member));

  for (i = 0; i < totalCount; i++) {
	 tbl->GetCell(cell,columnIndex,parentList[2*i]);
         f[i].name = (char *) malloc((cell.size()+1)*sizeof(char));
         strcpy(f[i].name,cell.c_str());
	 cell.clear();
    f[i].generation = parentList[2*i + 1];
  }
  
  if (parentList) free(parentList);
  numD = totalCount;
  return f;
}

string CifInterface::GetItemMandatoryCodeInherit(const string& dataBlockName,
                                         const string& categoryName,
			                 const string& itemKeyword,
                                         const Family decendents, 
                                         int decendentsNum)
/* -------------------------------------------------------------------------
   GetItemMandatoryCodeInherit(): returns the mandatory code for the item
   pointed to by dataBlockName, categoryName, and itemKeyword.  If one cannot
   be found, the Family decendents will be searched.  The number of decendents
   must also be specified by decendentsNum.
   ------------------------------------------------------------------------- */
{
  string ret;
  int i;
  string category;
  string keyword;

  if (dataBlockName.empty() || categoryName.empty() || itemKeyword.empty()) return ret;
  if ((decendentsNum < 0) || ((decendentsNum > 0) && (!decendents)))
    decendentsNum = 0;

  ret = GetItemMandatoryCode(dataBlockName, categoryName, itemKeyword); 
  if (!ret.empty()) return ret;
  else {
    for (i = 0; i < decendentsNum; i++) {
      category = GetCategoryByItemName(decendents[i].name);
      keyword  = GetItemKeywordByName(decendents[i].name);
      ret = GetItemMandatoryCode(dataBlockName, category, keyword);
      if (!ret.empty()) break;
    }
  return ret;
  }
}

string CifInterface::GetItemMandatoryCode(const string& dataBlockName,
                                         const string& categoryName,
			                 const string& itemKeyword)
/* -------------------------------------------------------------------------
   GetItemMandatoryCode():  returns the mandatory code for the item pointed to
   by dataBlockName, categoryName, and itemKeyword.
   ------------------------------------------------------------------------- */
{
  vector<string> ColumnNames;
  vector<string> values;
  vector<unsigned int> listOut;
  string ret;
  int columnIndex;
  int errCode;
  string cell;

  ISTable *tbl = _dic->GetTablePtr(dataBlockName, CIF_DDL_CATEGORY_ITEM);
  if(tbl == NULL) return ret;
  ColumnNames.push_back(CIF_DDL_ITEM_NAME);
  string name = GetItemName(categoryName, itemKeyword);
  values.push_back(name);    

  columnIndex = tbl->GetColumnIndex(CIF_DDL_ITEM_NAME);
  tbl->SetFlags(ISTable::DT_STRING | ISTable::CASE_INSENSE,columnIndex);

  tbl->Search(values,ColumnNames,listOut,errCode);
  if (errCode<0) return ret;
  else {
	 columnIndex = tbl->GetColumnIndex(CIF_DDL_ITEM_MANDATORY_CODE);
	 tbl->GetCell(cell,columnIndex,listOut[0]);
	 return cell;
  }

}

string CifInterface::GetItemDefaultValueInherit(const string& dataBlockName,
                                         const string& categoryName,
			                 const string& itemKeyword,
                                         const Family decendents, 
                                         int decendentsNum)
/* -------------------------------------------------------------------------
   GetItemDefaultInherit(): returns the default value for the item pointed to
   by dataBlockName, categoryName, and itemKeyword.  If one cannot be found,
   the Family decendents will be searched.  The number of decendents must also
   be specified by decendentsNum.
   ------------------------------------------------------------------------- */
{
  string ret;
  int i;
  string category;
  string keyword;

  if (dataBlockName.empty() || categoryName.empty() || itemKeyword.empty()) return ret;
  if ((decendentsNum < 0) || ((decendentsNum > 0) && (!decendents)))
    decendentsNum = 0;

  ret = GetItemDefaultValue(dataBlockName, categoryName, itemKeyword); 
  if (!ret.empty()) return ret;
  else {
    for (i = 0; i < decendentsNum; i++) {
      category = GetCategoryByItemName(decendents[i].name);
      keyword  = GetItemKeywordByName(decendents[i].name);
      ret = GetItemDefaultValue(dataBlockName, category, keyword);
      if (!ret.empty()) break;
    }
  return ret;
  }
}

string CifInterface::GetItemDefaultValue(const string& dataBlockName,
                                         const string& categoryName,
			                 const string& itemKeyword)
/* -------------------------------------------------------------------------
   GetItemDefaultValue(): returns the default value for the item
   pointed to by dataBlockName, categoryName, and itemKeyword.
   ------------------------------------------------------------------------- */
{
  vector<string> ColumnNames;
  vector<string> values;
  vector<unsigned int> listOut;
  string ret;
  int columnIndex;
  int errCode;
  string cell;

  ISTable *tbl = _dic->GetTablePtr(dataBlockName, CIF_DDL_CATEGORY_ITEM_DEFAULT);
  if(tbl == NULL) return ret;
  ColumnNames.push_back(CIF_DDL_ITEM_NAME);
  string name = GetItemName(categoryName, itemKeyword);
  values.push_back(name);    

  columnIndex = tbl->GetColumnIndex(CIF_DDL_ITEM_NAME);
  tbl->SetFlags(ISTable::DT_STRING | ISTable::CASE_INSENSE,columnIndex);

  tbl->Search(values,ColumnNames, listOut,errCode);
  if (errCode<0) return ret;
  else {
	 columnIndex = tbl->GetColumnIndex(CIF_DDL_ITEM_VALUE);
	 tbl->GetCell(cell,columnIndex,listOut[0]);
	 return cell;
  }

}

string CifInterface::GetItemUnitsInherit(const string& dataBlockName,
                                         const string& categoryName,
			                 const string& itemKeyword,
                                         const Family decendents, 
                                         int decendentsNum)
/* -------------------------------------------------------------------------
   GetItemUnitsInherit(): returns the units for the item pointed to by
   dataBlockName, categoryName, and itemKeyword.  If one cannot be found, the\    Family decendents will be searched.  The number of decendents must also
   be specified by decendentsNum.
   ------------------------------------------------------------------------- */
{
  string ret;
  int i;
  string category;
  string keyword;

  if (dataBlockName.empty() || categoryName.empty() || itemKeyword.empty()) return ret;
  if ((decendentsNum < 0) || ((decendentsNum > 0) && (!decendents)))
    decendentsNum = 0;

  ret = GetItemUnits(dataBlockName, categoryName, itemKeyword); 
  if (!ret.empty()) return ret;
  else {
    for (i = 0; i < decendentsNum; i++) {
      category = GetCategoryByItemName(decendents[i].name);
      keyword  = GetItemKeywordByName(decendents[i].name);
      ret = GetItemUnits(dataBlockName, category, keyword);
      if (!ret.empty()) break;
    }
  return ret;
  }
}

string CifInterface::GetItemUnits(const string& dataBlockName,
                                         const string& categoryName,
			                 const string& itemKeyword)
/* -------------------------------------------------------------------------
   GetItemUnits(): returns the default value for the item pointed to
   by dataBlockName, categoryName, and itemKeyword.
   ------------------------------------------------------------------------- */
{
  vector<string> ColumnNames;
  vector<string> values;
  vector<unsigned int> listOut;
  string ret;
  int columnIndex;
  int errCode;
  string cell;

  ISTable *tbl = _dic->GetTablePtr(dataBlockName, CIF_DDL_CATEGORY_ITEM_UNITS);
  if(tbl == NULL) return ret;
  ColumnNames.push_back(CIF_DDL_ITEM_NAME);
  string name = GetItemName(categoryName, itemKeyword);
  values.push_back(name);    

  columnIndex = tbl->GetColumnIndex(CIF_DDL_ITEM_NAME);
  tbl->SetFlags(ISTable::DT_STRING | ISTable::CASE_INSENSE,columnIndex);

  tbl->Search(values,ColumnNames, listOut,errCode);
  if (errCode<0) return ret;
  else {
	 columnIndex = tbl->GetColumnIndex(CIF_DDL_ITEM_CODE);
	 tbl->GetCell(cell,columnIndex,listOut[0]);
	 return cell;
  }

}

void CifInterface::GetItemDependentsInherit(vector<string>& itemDepInherit,
  const string& dataBlockName, const string& categoryName,
  const string& itemKeyword, const Family decendents, int decendentsNum)
/* -------------------------------------------------------------------------
   GetItemDependentsInherit(): returns the list of dependents for the item
   pointed to by dataBlockName, categoryName, and itemKeyword.  If one cannot
   be found, the Family decendents will be searched.  The number of decendents
   must also be specified by decendentsNum.
   -------------------------------------------------------------------------*/ 
{
  itemDepInherit.clear();

  int i;
  string category;
  string keyword;

  if (dataBlockName.empty() || categoryName.empty() || itemKeyword.empty()) return;
  if ((decendentsNum < 0) || ((decendentsNum > 0) && (!decendents)))
    decendentsNum = 0;

  GetItemDependents(itemDepInherit, dataBlockName, categoryName, itemKeyword); 
  if (!itemDepInherit.empty()) return;
  else {
    for (i = 0; i < decendentsNum; i++) {
      category = GetCategoryByItemName(decendents[i].name);
      keyword  = GetItemKeywordByName(decendents[i].name);
      GetItemDependents(itemDepInherit, dataBlockName, category, keyword);
      if (!itemDepInherit.empty()) break;
    }
  return;
  }
}

void CifInterface::GetItemDependents(vector<string>& itemDep,
  const string& dataBlockName, const string& categoryName,
  const string& itemKeyword)
/* -------------------------------------------------------------------------
   GetItemDependents(): returns the list of dependents for the item pointed to
   by dataBlockName, categoryName, and itemKeyword.
   -------------------------------------------------------------------------*/ 
{
  itemDep.clear();
  vector<string> ColumnNames;
  vector<string> values;
  vector<unsigned int> listOut;
  int columnIndex;
  int errCode;
  string cell;

  ISTable *tbl = _dic->GetTablePtr(dataBlockName, CIF_DDL_CATEGORY_ITEM_DEPENDENT);
  if(tbl == NULL) return;
  ColumnNames.push_back(CIF_DDL_ITEM_NAME);
  string name = GetItemName(categoryName, itemKeyword);
  values.push_back(name);    

  columnIndex = tbl->GetColumnIndex(CIF_DDL_ITEM_NAME);
  tbl->SetFlags(ISTable::DT_STRING | ISTable::CASE_INSENSE,columnIndex);

  tbl->Search(values,ColumnNames, listOut,errCode);
  if (errCode<0) return;
  else {
	 columnIndex = tbl->GetColumnIndex(CIF_DDL_ITEM_DEPENDENT_NAME);
	 int len = listOut.size();
	 for(int i=0; i<len; i++) {
		tbl->GetCell(cell,columnIndex,listOut[i]);
		itemDep.push_back(cell);
		cell.clear();
	 }
	 return;
  }
}

void CifInterface::GetItemSubcategoriesInherit(vector<string>& itemSubcategoriesInherit, const string& dataBlockName, const string& categoryName,
  const string& itemKeyword, const Family decendents, int decendentsNum)
/* -------------------------------------------------------------------------
   GetItemSubcategoriesInherit(): returns the list of subcategories to which
   the item pointed to by dataBlockName, categoryName, and itemKeyword
   belongs.  If none can be found, the Family decendents will be searched. The
   number of decendents must also be specified by decendentsNum and the number
   of subcategories will be returned via the reference num.
   -------------------------------------------------------------------------*/ 
{
  itemSubcategoriesInherit.clear();
  int i;
  string category;
  string keyword;

  if (dataBlockName.empty() || categoryName.empty() || itemKeyword.empty()) return;
  if ((decendentsNum < 0) || ((decendentsNum > 0) && (!decendents)))
    decendentsNum = 0;

  GetItemSubcategories(itemSubcategoriesInherit, dataBlockName, categoryName, itemKeyword); 
  if (!itemSubcategoriesInherit.empty()) return;
  else {
    for (i = 0; i < decendentsNum; i++) {
      category = GetCategoryByItemName(decendents[i].name);
      keyword  = GetItemKeywordByName(decendents[i].name);
      GetItemSubcategories(itemSubcategoriesInherit, dataBlockName, category, keyword);
      if (!itemSubcategoriesInherit.empty()) break;
    }
  return;
  }
}

void CifInterface::GetItemSubcategories(vector<string>& itemSubcategories,
  const string& dataBlockName, const string& categoryName,
  const string& itemKeyword)
/* -------------------------------------------------------------------------
   GetItemSubcategories(): returns the list of subcategories to which the item
   pointed to by dataBlockName, categoryName, and itemKeyword belong. The
   number of subcategories will be returned in the reference num.
   -------------------------------------------------------------------------*/ 
{
  itemSubcategories.clear();
  vector<string> ColumnNames;
  vector<string> values;
  vector<unsigned int> listOut;
  int columnIndex;
  int errCode;
  string cell;

  ISTable *tbl = _dic->GetTablePtr(dataBlockName, CIF_DDL_CATEGORY_ITEM_SUB_CATEGORY);
  if(tbl == NULL) return;
  ColumnNames.push_back(CIF_DDL_ITEM_NAME);
  string name = GetItemName(categoryName, itemKeyword);
  values.push_back(name);    

  columnIndex = tbl->GetColumnIndex(CIF_DDL_ITEM_NAME);
  tbl->SetFlags(ISTable::DT_STRING | ISTable::CASE_INSENSE,columnIndex);

  tbl->Search(values,ColumnNames,listOut,errCode);
  if (errCode<0) return;
  else {
	 columnIndex = tbl->GetColumnIndex(CIF_DDL_ITEM_ID);
	 int len = listOut.size();
	 for(int i=0; i<len; i++) {
		tbl->GetCell(cell,columnIndex,listOut[i]);
		itemSubcategories.push_back(cell);
		cell.clear();
	 }
	 return;
  }


}

void CifInterface::GetItemChildren(vector<string>& itemChildren,
  const string& dataBlockName, const string& categoryName,
  const string& itemKeyword)
/* -------------------------------------------------------------------------
   GetItemChildren(): returns the immediate children of the item pointed to by
   dataBlockName, categoryName, and itemKeyword. The number of children is
   returned in the reference num.
   -------------------------------------------------------------------------*/ 
{
  itemChildren.clear();

  vector<string> ColumnNames;
  vector<string> values;
  vector<unsigned int> listOut;
  int columnIndex;
  int errCode;
  string cell;

  ISTable *tbl = _dic->GetTablePtr(dataBlockName, CIF_DDL_CATEGORY_ITEM_LINKED);
  if(tbl == NULL) return;
  ColumnNames.push_back(CIF_DDL_ITEM_PARENT_NAME);
  string name = GetItemName(categoryName, itemKeyword);
  values.push_back(name);    

  columnIndex = tbl->GetColumnIndex(CIF_DDL_ITEM_PARENT_NAME);
  tbl->SetFlags(ISTable::DT_STRING | ISTable::CASE_INSENSE,columnIndex);

  tbl->Search(values,ColumnNames, listOut,errCode);
  if (errCode<0) return;
  else {
	 columnIndex = tbl->GetColumnIndex(CIF_DDL_ITEM_CHILD_NAME);
	 int len = listOut.size();
	 for(int i=0; i<len; i++) {
		tbl->GetCell(cell,columnIndex,listOut[i]);
		itemChildren.push_back(cell);
		cell.clear();
	 }
	 return;
  }
}

void CifInterface::GetItemParents(vector<string>& itemParents,
  const string& dataBlockName, const string& categoryName,
  const string& itemKeyword)
/* -------------------------------------------------------------------------
   GetItemParents(): returns the immediate parents of the item pointed to by
   dataBlockName, categoryName, and itemKeyword. The number of children is
   returned in the reference num.
   -------------------------------------------------------------------------*/ 
{
  itemParents.clear();

  vector<string> ColumnNames;
  vector<string> values;
  vector<unsigned int> listOut;
  int columnIndex;
  int errCode;
  string cell;

  ISTable *tbl = _dic->GetTablePtr(dataBlockName, CIF_DDL_CATEGORY_ITEM_LINKED);
  if(tbl == NULL) return;
  ColumnNames.push_back(CIF_DDL_ITEM_CHILD_NAME);
  string name = GetItemName(categoryName, itemKeyword);
  values.push_back(name);    

  columnIndex = tbl->GetColumnIndex(CIF_DDL_ITEM_CHILD_NAME);
  tbl->SetFlags(ISTable::DT_STRING | ISTable::CASE_INSENSE,columnIndex);

  tbl->Search(values,ColumnNames, listOut,errCode);
  if (errCode<0) return;
  else {
	 columnIndex = tbl->GetColumnIndex(CIF_DDL_ITEM_PARENT_NAME);
	 int len = listOut.size();
	 for(int i=0; i<len; i++) {
		tbl->GetCell(cell,columnIndex,listOut[i]);
		itemParents.push_back(cell);
		cell.clear();
	 }
	 return;
  }
}

void CifInterface::GetItemMethods(vector<string>& itemMethods,
  const string& dataBlockName, const string& categoryName,
  const string& itemKeyword)
/* -------------------------------------------------------------------------
   GetItemMethods(): returns a list of methods associated with the item
   pointed to by dataBlockName, categoryName, and itemKeyword. The number of 
   methods will be returned in the reference num.
   -------------------------------------------------------------------------*/ 
{
  vector<string> ColumnNames;
  vector<string> values;
  vector<unsigned int> listOut;
  int columnIndex;
  int errCode;
  string cell;

  itemMethods.clear();

  ISTable *tbl = _dic->GetTablePtr(dataBlockName, CIF_DDL_CATEGORY_ITEM_METHODS);
  if(tbl == NULL) return;
  ColumnNames.push_back(CIF_DDL_ITEM_NAME);
  string name = GetItemName(categoryName, itemKeyword);
  values.push_back(name);    

  columnIndex = tbl->GetColumnIndex(CIF_DDL_ITEM_NAME);
  tbl->SetFlags(ISTable::DT_STRING | ISTable::CASE_INSENSE,columnIndex);

  tbl->Search(values,ColumnNames, listOut, errCode);
  if (errCode<0) return;
  else {
	 columnIndex = tbl->GetColumnIndex(CIF_DDL_ITEM_METHOD_ID);
	 int len = listOut.size();
	 for(int i=0; i<len; i++) {
		tbl->GetCell(cell,columnIndex,listOut[i]);
		itemMethods.push_back(cell);
		cell.clear();
	 }
	 return;
  }
}

void CifInterface::GetItemTypeConditions(vector<string>& itemTypeCond,
  const string& dataBlockName, const string& categoryName,
  const string& itemKeyword)
/* -------------------------------------------------------------------------
   GetItemTypeConditions(): returns a list of type conditions associated with
   the item pointed to by dataBlockName, categoryName, and itemKeyword.  The
   number of type conditions is returned in the reference num.
   -------------------------------------------------------------------------*/ 
{
  itemTypeCond.clear();

  vector<string> ColumnNames;
  vector<string> values;
  vector<unsigned int> listOut;
  int columnIndex;
  int errCode;
  string cell;

  ISTable *tbl = _dic->GetTablePtr(dataBlockName, CIF_DDL_CATEGORY_ITEM_TYPE_CONDITIONS);
  if(tbl == NULL) return;
  ColumnNames.push_back(CIF_DDL_ITEM_NAME);
  string name = GetItemName(categoryName, itemKeyword);
  values.push_back(name);    

  columnIndex = tbl->GetColumnIndex(CIF_DDL_ITEM_NAME);
  tbl->SetFlags(ISTable::DT_STRING | ISTable::CASE_INSENSE,columnIndex);

  tbl->Search(values,ColumnNames, listOut, errCode);
  if (errCode<0) return;
  else {
	 columnIndex = tbl->GetColumnIndex(CIF_DDL_ITEM_CODE);
	 int len = listOut.size();
	 for(int i=0; i<len; i++) {
		tbl->GetCell(cell,columnIndex,listOut[i]);
		itemTypeCond.push_back(cell);
		cell.clear();
	 }
	 return;
  }
}

string CifInterface::GetCategoryDescription(const string& dataBlockName,
                                            const string& categoryName)
/*
-------------------------------------------------------------------------------
  GetCategoryDescription(): returns the description of the category pointed to
  by dataBlockName and categoryName.
-------------------------------------------------------------------------------
*/
{
  vector<string> ColumnNames;
  vector<string> values;
  vector<unsigned int> listOut;
  string ret;
  int columnIndex;
  int errCode;
  string cell;

  ISTable *tbl = _dic->GetTablePtr(dataBlockName, CIF_DDL_CATEGORY_CATEGORY);
  if(tbl == NULL) return NULL;
  ColumnNames.push_back(CIF_DDL_ITEM_ID);
  values.push_back(categoryName);

  columnIndex = tbl->GetColumnIndex(CIF_DDL_ITEM_ID);
  tbl->SetFlags(ISTable::DT_STRING | ISTable::CASE_INSENSE,columnIndex);

  tbl->Search(values,ColumnNames, listOut,errCode);
  if (errCode<0) return ret;
  else {
  columnIndex = tbl->GetColumnIndex(CIF_DDL_ITEM_DESCRIPTION);
	 tbl->GetCell(cell,columnIndex,listOut[0]);
	 ret = cell;
	 return ret;
  }
}

string CifInterface::GetCategoryDescriptionNDB(const string& dataBlockName,
					       const string& categoryName)
/*
-------------------------------------------------------------------------------
  GetCategoryDescription(): returns the description of the category pointed to
  by dataBlockName and categoryName.
-------------------------------------------------------------------------------
*/
{
  vector<string> ColumnNames;
  vector<string> values;
  vector<unsigned int> listOut;
  string ret;
  int columnIndex;
  int errCode;
  string cell;

  ISTable *tbl = _dic->GetTablePtr(dataBlockName, CIF_DDL_CATEGORY_NDB_CATEGORY_DESCRIPTION);
  if(tbl == NULL) return ret;
  ColumnNames.push_back(CIF_DDL_ITEM_ID);
  values.push_back(categoryName);

  columnIndex = tbl->GetColumnIndex(CIF_DDL_ITEM_ID);
  tbl->SetFlags(ISTable::DT_STRING | ISTable::CASE_INSENSE,columnIndex);

  tbl->Search(values,ColumnNames, listOut, errCode);
  if (errCode<0) return ret;
  else {
  columnIndex = tbl->GetColumnIndex(CIF_DDL_ITEM_DESCRIPTION);
	 tbl->GetCell(cell,columnIndex,listOut[0]);
	 ret = cell;
	 return ret;
  }


}


string CifInterface::GetCategoryMandatoryCode(const string& dataBlockName,
                                              const string& categoryName)
/*
-------------------------------------------------------------------------------
  GetCategoryMandatoryCode(): returns the mandatory code of the category
  pointed to by dataBlockName and categoryName.
-------------------------------------------------------------------------------
*/
{
  string ret;

  vector<string> ColumnNames;
  vector<string> values;
  vector<unsigned int> listOut;
  int columnIndex;
  int errCode;
  string cell;

  if (dataBlockName.empty() || categoryName.empty()) {
    return ret;
  }

  ISTable * tbl = _dic->GetTablePtr(dataBlockName,CIF_DDL_CATEGORY_CATEGORY);
  if(tbl == NULL) return ret;
  ColumnNames.push_back(CIF_DDL_ITEM_ID);
  values.push_back(categoryName);

  columnIndex = tbl->GetColumnIndex(CIF_DDL_ITEM_ID);
  tbl->SetFlags(ISTable::DT_STRING | ISTable::CASE_INSENSE,columnIndex);

  tbl->Search(values, ColumnNames,listOut,errCode);
  if (errCode<0)  {
	 return ret;
  }
  else {
	 columnIndex=tbl->GetColumnIndex(CIF_DDL_ITEM_MANDATORY_CODE);
	 tbl->GetCell(cell,columnIndex,listOut[0]);
	 ret = cell;
	 return ret;
  }
}

void CifInterface::GetCategoryKeyItems(vector<string>& keyItems,
  const string& dataBlockName, const string& categoryName)
/* -------------------------------------------------------------------------
   GetCategoryKeyItems(): returns the keys of the category pointed to by
   dataBlockName and categoryName.  The number of keys will be returned in the
   reference num.
   -------------------------------------------------------------------------*/ 
{

  keyItems.clear();

  vector<string> ColumnNames;
  vector<string> values;
  vector<unsigned int> listOut;
  int columnIndex;
  int errCode;
  string cell;

  if (dataBlockName.empty() || categoryName.empty()) {
    return;
  }

  ISTable * tbl = _dic->GetTablePtr(dataBlockName,CIF_DDL_CATEGORY_CATEGORY_KEY);
  if(tbl == NULL) {
    return;
  }

  ColumnNames.push_back(CIF_DDL_ITEM_ID);
  values.push_back(categoryName);

  columnIndex = tbl->GetColumnIndex(CIF_DDL_ITEM_ID);
  tbl->SetFlags(ISTable::DT_STRING | ISTable::CASE_INSENSE,columnIndex);

  tbl->Search(values,ColumnNames, listOut, errCode);
  if (errCode<0)  {
	 return;
  }
  else {
	 columnIndex=tbl->GetColumnIndex(CIF_DDL_ITEM_NAME);
	 int len = listOut.size();
	 for(int i=0; i<len; i++) {
		tbl->GetCell(cell,columnIndex,listOut[i]);
		keyItems.push_back(cell);
		cell.clear();
	 }
	 return;
  }
}

void CifInterface::GetCategoryCategoryGroups(vector<string>& categoryCategoryGroups, const string& dataBlockName, const string& categoryName)
/* -------------------------------------------------------------------------
   GetCategoryCategoryGroups(): returns a list of the category groups to which
   the category pointed to by dataBlockName and categoryName belongs.
   -------------------------------------------------------------------------*/ 
{
  categoryCategoryGroups.clear();

  vector<string> ColumnNames;
  vector<string> values;
  vector<unsigned int> listOut;
  int columnIndex;
  int errCode;
  string cell;

  if (dataBlockName.empty() || categoryName.empty()) {
    return;
  }

  ISTable * tbl = _dic->GetTablePtr(dataBlockName,CIF_DDL_CATEGORY_CATEGORY_GROUP);
  if(tbl == NULL) return;
  ColumnNames.push_back(CIF_DDL_ITEM_CATEGORY_ID);
  values.push_back(categoryName);

  columnIndex = tbl->GetColumnIndex(CIF_DDL_ITEM_CATEGORY_ID);
  tbl->SetFlags(ISTable::DT_STRING | ISTable::CASE_INSENSE,columnIndex);

  tbl->Search(values,ColumnNames, listOut, errCode);
  if (errCode<0)  {
	 return;
  }
  else {
	 columnIndex=tbl->GetColumnIndex(CIF_DDL_ITEM_ID);
	 int len = listOut.size();
	 for(int i=0; i<len; i++) {
		tbl->GetCell(cell,columnIndex,listOut[i]);
		categoryCategoryGroups.push_back(cell);
		cell.clear();
	 }
	 return;
  }
}

void CifInterface::GetCategoryCategoryMethods(vector<string>& categoryCategoryMethods, const string& dataBlockName, const string& categoryName)
/* -------------------------------------------------------------------------
   GetCategoryCategoryMethods(): returns a list of the category methods
   associated with the category pointed to by dataBlockName and categoryName.
   The number of such methods is returned in the reference num.
   -------------------------------------------------------------------------*/ 
{
  categoryCategoryMethods.clear();
  vector<string> ColumnNames;
  vector<string> values;
  vector<unsigned int> listOut;
  int columnIndex;
  int errCode;
  string cell;

  if (dataBlockName.empty() || categoryName.empty()) {
    return;
  }

  ISTable * tbl = _dic->GetTablePtr(dataBlockName,CIF_DDL_CATEGORY_CATEGORY_METHODS);
  if(tbl == NULL) return;
  ColumnNames.push_back(CIF_DDL_ITEM_CATEGORY_ID);
  values.push_back(categoryName);

  columnIndex = tbl->GetColumnIndex(CIF_DDL_ITEM_CATEGORY_ID);
  tbl->SetFlags(ISTable::DT_STRING | ISTable::CASE_INSENSE,columnIndex);

  tbl->Search(values,ColumnNames, listOut, errCode);
  if (errCode<0)  {
	 return;
  }
  else {
	 columnIndex=tbl->GetColumnIndex(CIF_DDL_ITEM_METHOD_ID);
	 for(int i=0; i<(int)listOut.size(); i++) {
		tbl->GetCell(cell,columnIndex,listOut[i]);
		categoryCategoryMethods.push_back(cell);
		cell.clear();
	 }
	 return;
  }
}

string CifInterface::GetSubCategoryDescription(const string& dataBlockName,
                                               const string& subcategoryName)
/* ----------------------------------------------------------------------------
   GetSubCategoryDescription(): returns the description of the subcategory
   pointed to by dataBlockName and subcategoryName.
   --------------------------------------------------------------------------*/
{
  string ret;
  vector<string> ColumnNames;
  vector<string> values;
  vector<unsigned int> listOut;
  int columnIndex;
  int errCode;
  string cell;

  if (dataBlockName.empty() || subcategoryName.empty()) {
    return ret;
  }

  ISTable * tbl = _dic->GetTablePtr(dataBlockName,CIF_DDL_CATEGORY_SUB_CATEGORY);
  if(tbl == NULL) return ret;
  ColumnNames.push_back(CIF_DDL_ITEM_ID);
  values.push_back(subcategoryName);

  columnIndex = tbl->GetColumnIndex(CIF_DDL_ITEM_ID);
  tbl->SetFlags(ISTable::DT_STRING | ISTable::CASE_INSENSE,columnIndex);

  tbl->Search(values,ColumnNames, listOut, errCode);
  if (errCode<0)  {
	 return ret;
  }
  else {
	 columnIndex=tbl->GetColumnIndex(CIF_DDL_ITEM_DESCRIPTION);
	 tbl->GetCell(cell,columnIndex,listOut[0]);
	 return cell;
  }
}



void CifInterface::GetSubCategoryMethods(vector<string>& subcategoryMethods,
  const string& dataBlockName, const string& subcategoryName)
/* -------------------------------------------------------------------------
   GetSubCategoryMethods(): returns a list of the methods associated with the
   subcategory pointed to by dataBlockName and subcategoryName.  The number of
   such methods is returned in the reference num.
   -------------------------------------------------------------------------*/ 
{
  subcategoryMethods.clear();

  vector<string> ColumnNames;
  vector<string> values;
  vector<unsigned int> listOut;
  int columnIndex;
  int errCode;
  string cell;

  if (dataBlockName.empty() || subcategoryName.empty()) {
    return;
  }

  ISTable * tbl = _dic->GetTablePtr(dataBlockName,CIF_DDL_CATEGORY_SUB_CATEGORY_METHODS);
  if(tbl == NULL) return;
  ColumnNames.push_back(CIF_DDL_ITEM_SUB_CATEGORY_ID);
  values.push_back(subcategoryName);

  columnIndex = tbl->GetColumnIndex(CIF_DDL_ITEM_SUB_CATEGORY_ID);
  tbl->SetFlags(ISTable::DT_STRING | ISTable::CASE_INSENSE,columnIndex);

  tbl->Search(values,ColumnNames, listOut, errCode);
  if (errCode<0)  {
	 return;
  }
  else {
	 columnIndex=tbl->GetColumnIndex(CIF_DDL_ITEM_METHOD_ID);
	 for(int i=0; i<(int)listOut.size(); i++) {
		tbl->GetCell(cell,columnIndex,listOut[i]);
		subcategoryMethods.push_back(cell);
		cell.clear();
	 }
	 return;
  }
}

string CifInterface::GetDictionaryTitle(const string& dataBlockName)
/* ----------------------------------------------------------------------------
   GetDictionaryTitle(): returns the title of the dictionary with the datablock
   name dataBlockName.  These are not necessarily the same, although they may
   be.
   --------------------------------------------------------------------------*/
{
  string ret;
  vector<string> ColumnNames;
  vector<string> values;
  vector<unsigned int> listOut;
  int columnIndex;
  int errCode;
  string cell;

  if (dataBlockName.empty()) {
    return ret;
  }

  ISTable * tbl = _dic->GetTablePtr(dataBlockName,CIF_DDL_CATEGORY_DICTIONARY);
  if(tbl == NULL) return ret;
  ColumnNames.push_back(CIF_DDL_ITEM_DATABLOCK_ID);
  values.push_back(dataBlockName);

  columnIndex = tbl->GetColumnIndex(CIF_DDL_ITEM_DATABLOCK_ID);
  tbl->SetFlags(ISTable::DT_STRING | ISTable::CASE_INSENSE,columnIndex);

  tbl->Search(values,ColumnNames, listOut, errCode);
  if (errCode<0)  {
	 return ret;
  }
  else {
	 columnIndex=tbl->GetColumnIndex(CIF_DDL_ITEM_TITLE);
	 tbl->GetCell(cell,columnIndex,listOut[0]);
	 ret = cell;
	 return ret;
  }

}
  
string CifInterface::GetDictionaryVersion(const string& dataBlockName)
/* ----------------------------------------------------------------------------
   GetDictionaryVersion(): returns the version of the dictionary with the
   datablock name dataBlockName.
   --------------------------------------------------------------------------*/
{
  string ret;
  vector<string> ColumnNames;
  vector<string> values;
  vector<unsigned int> listOut;
  int columnIndex;
  int errCode;
  string cell;

  if (dataBlockName.empty()) return ret;

  ISTable * tbl = _dic->GetTablePtr(dataBlockName,CIF_DDL_CATEGORY_DICTIONARY);
  if(tbl == NULL) return ret;
  ColumnNames.push_back(CIF_DDL_ITEM_DATABLOCK_ID);
  values.push_back(dataBlockName);

  columnIndex = tbl->GetColumnIndex(CIF_DDL_ITEM_DATABLOCK_ID);
  tbl->SetFlags(ISTable::DT_STRING | ISTable::CASE_INSENSE,columnIndex);

  tbl->Search(values,ColumnNames, listOut, errCode);
  if (errCode<0) return ret;
  else {
	 columnIndex=tbl->GetColumnIndex(CIF_DDL_ITEM_VERSION);
	 tbl->GetCell(cell,columnIndex,listOut[0]);
	 ret = cell;
	 return ret;
  }
}
  
void CifInterface::GetDictionaryCategories(vector<string>& dictSubcategories,
  const string& dataBlockName)
/* ----------------------------------------------------------------------------
   GetDictionaryCategories(): returns a list of the categories in the
   dictionary with the datablock name dataBlockName.  The number of 
	categories is returned in the reference num.
   --------------------------------------------------------------------------*/
{
  dictSubcategories.clear();
  int columnIndex;
  vector<string> column;
  vector<string> list;
  
  if (dataBlockName.empty()) return;

  ISTable * tbl = _dic->GetTablePtr(dataBlockName,CIF_DDL_CATEGORY_CATEGORY);
  if(tbl == NULL) return;
  columnIndex=tbl->GetColumnIndex(CIF_DDL_ITEM_ID);
  list.push_back(CIF_DDL_ITEM_ID);
  tbl->CreateIndex("index0",list);
  tbl->GetColumn(column, columnIndex,"index0");

  for(int i=0; i<(int)column.size(); i++) {
	 dictSubcategories.push_back(column[i]);
  }
  column.clear();
  return;
}

void CifInterface::GetDictionarySubcategories(vector<string>& dictSubcategories, const string& dataBlockName)
/* ----------------------------------------------------------------------------
   GetDictionarySubcategories(): returns a list of subcategories in the
   dictionary wit the datablock name dataBlockName.  The number of 
	subcategories is returned in the reference num. 
   --------------------------------------------------------------------------*/
{
  dictSubcategories.clear();

  int columnIndex;
  vector<string> column;
  vector<string> list;

  if (dataBlockName.empty()) {
    return;
  }

  ISTable * tbl = _dic->GetTablePtr(dataBlockName,CIF_DDL_CATEGORY_SUB_CATEGORY);
  if(tbl == NULL) return;
  columnIndex=tbl->GetColumnIndex(CIF_DDL_ITEM_ID);
  list.push_back(CIF_DDL_ITEM_ID);
  tbl->CreateIndex("index0",list);
  tbl->GetColumn(column, columnIndex,"index0");

  for(int i=0; i<(int)column.size(); i++) {
	 dictSubcategories.push_back(column[i]);
  }
  column.clear();
  return;
}

void CifInterface::GetDictionaryItems(vector<string>& dictItems,
  const string& dataBlockName)
/* ----------------------------------------------------------------------------
   GetDictionaryItems(): returns a list of the items in the dictionary with the
   datablock name datablockname.  The number of items is returned in the 
	reference num.
   --------------------------------------------------------------------------*/
{
  dictItems.clear();
  int columnIndex;
  vector<string> column;
  vector<string> list;

  if (dataBlockName.empty()) {
    return;
  }

  ISTable * tbl = _dic->GetTablePtr(dataBlockName,CIF_DDL_CATEGORY_ITEM);
  if(tbl == NULL) return;
  list.push_back(CIF_DDL_ITEM_NAME);
  tbl->CreateIndex("index0",list);
  columnIndex=tbl->GetColumnIndex(CIF_DDL_ITEM_NAME);
  tbl->GetColumn(column, columnIndex,"index0");

  for(int i=0; i<(int)column.size(); i++) {
	 dictItems.push_back(column[i]);
  }
  column.clear();
  return;
}

string CifInterface::GetDictionaryDescription(string& dataBlockName)
/* ----------------------------------------------------------------------------
   GetDictionaryDescription(): returns the description of the dictionary with
   the datablock name dataBlockName.
   --------------------------------------------------------------------------*/
{
  string ret;
  vector<string> ColumnNames;
  vector<string> values;
  vector<unsigned int> listOut;
  int columnIndex;
  int errCode;
  string cell;

  if (dataBlockName.empty()) {
    return ret;
  }

  ISTable * tbl = _dic->GetTablePtr(dataBlockName,CIF_DDL_CATEGORY_DATABLOCK);
  if(tbl == NULL) return ret;
  ColumnNames.push_back(CIF_DDL_ITEM_ID);
  values.push_back(dataBlockName);

  columnIndex = tbl->GetColumnIndex(CIF_DDL_ITEM_ID);
  tbl->SetFlags(ISTable::DT_STRING | ISTable::CASE_INSENSE,columnIndex);

  tbl->Search(values,ColumnNames,listOut,errCode);
  if (errCode<0) return ret;
  else {
	 columnIndex=tbl->GetColumnIndex(CIF_DDL_ITEM_DESCRIPTION);
	 tbl->GetCell(cell,columnIndex,listOut[0]);
	 ret = cell;
	 return ret;
  }
}


// Private Section

void CifInterface::Clear()
/* ----------------------------------------------------------------------------
   Clear(): frees up all memory used by the object
   --------------------------------------------------------------------------*/
{
  _ddl->Clear();
//  _ddlformat->Clear();
  _dic->Clear();
//  _dicformat->Clear();
  verify = 1;
}

void CifInterface::Reset()
/* ----------------------------------------------------------------------------
   Reset(): same as Clear() except that the filename is no longer associated
   with the object.
   --------------------------------------------------------------------------*/
{
  Clear();
  filename.clear();
}



// other helper functions

int sCompareSort(const void * a, const void * b)
/* ----------------------------------------------------------------------------
   sCompareSort(): (string compare) is a function used by qsort() to sort an
   array of char *.
   --------------------------------------------------------------------------*/
{
  // VLAD: FIX THIS AS PART OF CONTAINER FIX
  return strcmp(*((char **) a), *((char **) b)); 
}

