/*
FILE:     CifLoadReorganizer.C
*/
/*
VERSION:  7.001
*/
/*
DATE:     5/18/2006
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2006 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
#include <string>
#include <string.h>
#include "CifLoadReorganizer.h"

CifInterface * ci = NULL; // used to load the objects
extern FileNavigator * fnav; // used to write the objects

//const int DictionaryLoadObj::_verbose = 1;
const int DictionaryLoadObj::DICTIONARY_FIELDS = 35;

CategoryLoadObj::CategoryLoadObj() {
/*
-------------------------------------------------------------------------------
  constructor that initializes the object and instantiates ci
-------------------------------------------------------------------------------
*/
  Clear();
  if (!ci) ci = new CifInterface;
}

CategoryLoadObj::CategoryLoadObj(const string& dataBlockName,
				 const string& categoryId) {
/*
-------------------------------------------------------------------------------
  constructor that initializes the object and sets the keys
-------------------------------------------------------------------------------
*/
  if (!ci) ci = new CifInterface;
  Clear();
  SetId(dataBlockName, categoryId);
}  

void CategoryLoadObj::SetId(const string& dataBlockName,
                                                   const string& categoryId) {
/*
-------------------------------------------------------------------------------
  SetId() sets the keys of the object
-------------------------------------------------------------------------------
*/
  _id = categoryId;  
  _datablock = dataBlockName;
}

int CategoryLoadObj::LoadObject() {
/*
-------------------------------------------------------------------------------
  LoadObject() will call all of the private Load _methods so that all data is
  retrieved from ciflib and stored in the object. Returns KEYS_NOT_SET upon
  failure.
-------------------------------------------------------------------------------
*/
  if (_datablock.empty() || _id.empty()) return KEYS_NOT_SET; 
  LoadDescription();
  LoadMandatoryCode();
  LoadKeyItems();
  LoadExamples();
  LoadCategoryGroups();
  LoadCategoryMethods();
  LoadDescriptionNDB();
  LoadExamplesNDB();
  return NO_ERROR;
}

/*
   The following Load _methods use ciflib (either directly or through the
   CifInterface class) to populate the data section of the object.
*/

void CategoryLoadObj::LoadDescription() {
  if (_id.empty() || _datablock.empty())
    _description.clear();
  else _description = ci->GetCategoryDescription(_datablock, _id);
}

void CategoryLoadObj::LoadDescriptionNDB() {
  if (_id.empty() || _datablock.empty())
    _descriptionNDB.clear();
  else _descriptionNDB = ci->GetCategoryDescriptionNDB(_datablock, _id);
}

void CategoryLoadObj::LoadMandatoryCode() {
  if (_id.empty() || _datablock.empty())
    _mandatoryCode.clear();
  else _mandatoryCode = ci->GetCategoryMandatoryCode(_datablock, _id);
}

void CategoryLoadObj::LoadKeyItems()
{
  ci->GetCategoryKeyItems(_keyItems, _datablock, _id);
}

void CategoryLoadObj::LoadExamples() {
  int i;
  ISTable * tbl;
  vector<string> ColumnNames;
  vector<string> values;
  vector<unsigned int> listOut;
  string cell;
  int errCode;
  int columnIndex1;
  int columnIndex2;

  if (_id.empty() || _datablock.empty()) {
    _examplesCase.clear();
    _examplesDetail.clear();
    return;
  }
  
  tbl = ci->GetTablePtr(_datablock, CIF_DDL_CATEGORY_CATEGORY_EXAMPLES);
  if (tbl == NULL) {
    _examplesCase.clear();
    _examplesDetail.clear();
    return;
  }
  ColumnNames.push_back(CIF_DDL_ITEM_ID);
  values.push_back(_id); 
  
  columnIndex1 = tbl->GetColumnIndex(CIF_DDL_ITEM_ID);
  tbl->SetFlags(ISTable::DT_STRING | ISTable::CASE_INSENSE,columnIndex1);

  tbl->Search(values,ColumnNames,listOut,errCode);
  if (errCode<0){
    _examplesCase.clear();
    _examplesDetail.clear();
    return;
  }
  
  columnIndex1 = tbl->GetColumnIndex(CIF_DDL_ITEM_CASE);
  columnIndex2 = tbl->GetColumnIndex(CIF_DDL_ITEM_DETAIL);
  for (i = 0; i < (int)listOut.size(); i++) {
	 tbl->GetCell(cell,columnIndex1,listOut[i]);
	 _examplesCase.push_back(cell);
	 cell.clear();
	 tbl->GetCell(cell,columnIndex2,listOut[i]);
	 _examplesDetail.push_back(cell);
	 cell.clear();
  }
  return;
}

void CategoryLoadObj::LoadExamplesNDB() {
  int i;
  ISTable * tbl;
  vector<string> ColumnNames;
  vector<string> values;
  vector<unsigned int> listOut;
  string cell;
  int errCode;
  int columnIndex1;
  int columnIndex2;

  if (_id.empty() || _datablock.empty()) {
    _examplesCaseNDB.clear();
    _examplesDetailNDB.clear();
    return;
  }
  
  tbl = ci->GetTablePtr(_datablock, CIF_DDL_CATEGORY_NDB_CATEGORY_EXAMPLES);
  if (tbl == NULL) {
    _examplesCaseNDB.clear();
    _examplesDetailNDB.clear();
    return;
  }
  ColumnNames.push_back(CIF_DDL_ITEM_ID);
  values.push_back(_id); 
  
  columnIndex1 = tbl->GetColumnIndex(CIF_DDL_ITEM_ID);
  tbl->SetFlags(ISTable::DT_STRING | ISTable::CASE_INSENSE,columnIndex1);

  tbl->Search(values,ColumnNames,listOut,errCode);
  if (errCode<0){
    _examplesCaseNDB.clear();
    _examplesDetailNDB.clear();
    return;
  }

  columnIndex1 = tbl->GetColumnIndex(CIF_DDL_ITEM_CASE);
  columnIndex2 = tbl->GetColumnIndex(CIF_DDL_ITEM_DETAIL);
  for (i = 0; i < (int)listOut.size(); i++) {
	 tbl->GetCell(cell,columnIndex1,listOut[i]);
	 _examplesCaseNDB.push_back(cell);
	 cell.clear();
	 tbl->GetCell(cell,columnIndex2,listOut[i]);
	 _examplesDetailNDB.push_back(cell);
	 cell.clear();
  }

}


void CategoryLoadObj::LoadCategoryGroups() {
  ci->GetCategoryCategoryGroups(_categoryGroups, _datablock, _id);
}

void CategoryLoadObj::LoadCategoryMethods() {
  ci->GetCategoryCategoryMethods(_categoryMethods, _datablock, _id);
}

int CategoryLoadObj::WriteObject() {
/*
------------------------------------------------------------------------------
  WriteObject() calls the private Write methods to write the whole object to
  persistent storage. It then tacks on an index into the file for purposes of
  reconstructing the object.
------------------------------------------------------------------------------
*/
  Word i;
  i = 0;
  if (_index) free(_index);
  _index = (Word *) malloc(CATEGORY_FIELDS*WORDSIZE);
  _index[0] = WriteDescription();
  _index[1] = WriteMandatoryCode();
  _index[2] = WriteKeyItems();
  _index[3] = WriteExamplesCase();
  _index[4] = WriteExamplesDetail();
  _index[5] = WriteCategoryGroups();
  _index[6] = WriteCategoryMethods();
  _index[7] = WriteDescriptionNDB();
  _index[8] = WriteExamplesCaseNDB();
  _index[9] = WriteExamplesDetailNDB();

  int err = fnav->WriteWords(_index, CATEGORY_FIELDS, i);
  if (err == NO_ERROR) 
    return i; 
  else return err;
}

/*
  The following Write _methods utilize the FileNavigator class to write the
  data of the object onto a persistent stroage file.
*/

int CategoryLoadObj::WriteDescription() {
  Word i;
  i = 0;
  int err;
  if (_description.empty()) return NO_ERROR;
  if ((err = fnav->WriteString(_description, i))==NO_ERROR) return i;
  else return err; 
}

int CategoryLoadObj::WriteDescriptionNDB() {
  Word i;
  i = 0;
  int err;
  if (_descriptionNDB.empty()) return NO_ERROR;
  if ((err = fnav->WriteString(_descriptionNDB, i))==NO_ERROR) return i;
  else return err; 
}

int CategoryLoadObj::WriteMandatoryCode() {
  Word i;
  i = 0;
  int err;
  if (_mandatoryCode.empty()) return NO_ERROR;
  if ((err = fnav->WriteString(_mandatoryCode, i))==NO_ERROR) return i;
  else return err; 
}

int CategoryLoadObj::WriteKeyItems() {
  Word i;
  i = 0;
  int err;
  if (_keyItems.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_keyItems, i))==NO_ERROR) return i;
  else return err; 
}

int CategoryLoadObj::WriteExamplesCase() {
  Word i;
  i = 0;
  int err;
  if (_examplesCase.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_examplesCase, i))==NO_ERROR) return i;
  else return err; 
}

int CategoryLoadObj::WriteExamplesDetail() {
  Word i;
  i = 0;
  int err;
  if (_examplesDetail.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_examplesDetail, i))==NO_ERROR) return i;
  else return err; 
}

int CategoryLoadObj::WriteExamplesCaseNDB() {
  Word i;
  i = 0;
  int err;
  if (_examplesCaseNDB.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_examplesCaseNDB, i))==NO_ERROR) return i;
  else return err; 
}

int CategoryLoadObj::WriteExamplesDetailNDB() {
  Word i;
  i = 0;
  int err;
  if (_examplesDetailNDB.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_examplesDetailNDB, i))==NO_ERROR) return i;
  else return err; 
}

int CategoryLoadObj::WriteCategoryGroups() {
  Word i;
  i = 0;
  int err;
  if (_categoryGroups.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_categoryGroups, i))==NO_ERROR) return i;
  else return err; 
}

int CategoryLoadObj::WriteCategoryMethods() {
  Word i;
  i = 0;
  int err;
  if (_categoryMethods.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_categoryMethods, i))==NO_ERROR) return i;
  else return err; 
}


ItemLoadObj::ItemLoadObj() {
/*
------------------------------------------------------------------------------
  constructor that initializes the object and instantiates ci
------------------------------------------------------------------------------
*/
  Clear();
  if (!ci) ci = new CifInterface;
}

ItemLoadObj::ItemLoadObj(const string& dataBlockName, const string& itemName) {
/*
------------------------------------------------------------------------------
  constructor that initializes the object and sets the keys
------------------------------------------------------------------------------
*/
  Clear();
  if (!ci) ci = new CifInterface;
  SetName(dataBlockName, itemName);
}

void ItemLoadObj::SetName(const string& dataBlockName, const string& itemName) {
/*
------------------------------------------------------------------------------
  SetName() sets the keys of the object
------------------------------------------------------------------------------
*/
  _name = itemName;
  _datablock = dataBlockName;

  _category = ci->GetCategoryByItemName(_name);
  _keyword = ci->GetItemKeywordByName(_name);

  _decendency = ci->GetItemDecendency(_datablock, _category, _keyword, _decendencyNum);
}

int ItemLoadObj::LoadObject() {
 /*
-------------------------------------------------------------------------------
  LoadObject() will call all of the private Load methods so that all data is
  retrieved from ciflib and stored in the object. Returns KEYS_NOT_SET upon
  failure.
-------------------------------------------------------------------------------
*/
 if (_datablock.empty() || _name.empty() || _category.empty() || _keyword.empty()) return KEYS_NOT_SET;
  LoadDescription();
  LoadType();
  LoadPrimitiveCode();
  LoadExamples();
  LoadEnumeration();
  LoadMandatoryCode();
  LoadDefaultValue();
  LoadUnits();
  LoadRanges();
  LoadAliases();
  LoadItemStructure();
  LoadDependents();
  LoadRelated();
  LoadSubcategories();
  LoadLinkedChildren();
  LoadLinkedParents();
  LoadMethods();
  LoadTypeConditions();
  LoadDescriptionNDB();
  LoadExamplesNDB();
  LoadEnumerationNDB();
  return NO_ERROR;
}

/*
   The following Load methods use ciflib (either directly or through the
   CifInterface class) to populate the data section of the object.
*/

void ItemLoadObj::LoadDescription() {
  if (_name.empty() || _datablock.empty() || _category.empty() || _keyword.empty()) {
    _description.clear();
    return;
  }
  _description = ci->GetItemDescriptionInherit(_datablock, _category, _keyword, _decendency, _decendencyNum);
}

void ItemLoadObj::LoadDescriptionNDB() {
  if (_name.empty() || _datablock.empty() || _category.empty() || _keyword.empty()) {
    _descriptionNDB.clear();
    return;
  }
  _descriptionNDB = ci->GetItemDescriptionInheritNDB(_datablock, _category, _keyword, _decendency, _decendencyNum);
}

void ItemLoadObj::LoadType() {
  if (_name.empty() || _datablock.empty() || _category.empty() || _keyword.empty()) {
    _type.clear();
    return;
  }
    _type = ci->GetItemDataTypeCodeInherit(_datablock, _category, _keyword, _decendency, _decendencyNum);
}

void ItemLoadObj::LoadPrimitiveCode() {
  ISTable * tbl;
  vector<string> ColumnNames;
  vector<string> values;
  vector<unsigned int> listOut;
  string cell;
  int errCode;
  int columnIndex1;
  int columnIndex2;

  if (_name.empty() || _datablock.empty() || _category.empty() || _keyword.empty() || _type.empty()) {
    _primitiveCode.clear();
    _regex.clear();
    return;
  }
  
  tbl = ci->GetTablePtr(_datablock, CIF_DDL_CATEGORY_ITEM_TYPE_LIST);
  if (tbl == NULL){
    _primitiveCode.clear();
    _regex.clear();
    return;
  }
  ColumnNames.push_back(CIF_DDL_ITEM_CODE);
  values.push_back(_type);  
  
  columnIndex1 = tbl->GetColumnIndex(CIF_DDL_ITEM_CODE);
  tbl->SetFlags(ISTable::DT_STRING | ISTable::CASE_INSENSE,columnIndex1);

  tbl->Search(values,ColumnNames,listOut,errCode);
  if (errCode<0) {
    _primitiveCode.clear();
    _regex.clear();
  }
  else {
	 columnIndex1 = tbl->GetColumnIndex(CIF_DDL_ITEM_PRIMITIVE_CODE);
	 columnIndex2 = tbl->GetColumnIndex(CIF_DDL_ITEM_CONSTRUCT);
	 
	 tbl->GetCell(cell,columnIndex1,listOut[0]);
	 _primitiveCode = cell;
	 cell.clear();
	 tbl->GetCell(cell,columnIndex2,listOut[0]);
	 _regex = cell;
  }
}

void ItemLoadObj::LoadExamples() {
  int i;
  ISTable * tbl;
  vector<string> ColumnNames;
  vector<string> values;
  vector<unsigned int> listOut;
  string cell;
  int errCode;
  int columnIndex1;
  int columnIndex2;

  if (_name.empty() || _datablock.empty() || _category.empty() || _keyword.empty()) {
    _examplesCase.clear();
    _examplesDetail.clear();
    return;
  }
  
  tbl = ci->GetTablePtr(_datablock, CIF_DDL_CATEGORY_ITEM_EXAMPLES);
  if (tbl == NULL) {
    _examplesCase.clear();
    _examplesDetail.clear();
    return;
  }
  ColumnNames.push_back(CIF_DDL_ITEM_NAME);
  values.push_back(_name); 
  
  columnIndex1 = tbl->GetColumnIndex(CIF_DDL_ITEM_NAME);
  tbl->SetFlags(ISTable::DT_STRING | ISTable::CASE_INSENSE,columnIndex1);

  tbl->Search(values,ColumnNames,listOut,errCode);
  if (errCode<0){
    _examplesCase.clear();
    _examplesDetail.clear();
    return;
  }
  columnIndex1 = tbl->GetColumnIndex(CIF_DDL_ITEM_CASE);
  columnIndex2 = tbl->GetColumnIndex(CIF_DDL_ITEM_DETAIL);
  for (i = 0; i < (int)listOut.size(); i++) {
	 tbl->GetCell(cell,columnIndex1,listOut[i]);
	 _examplesCase.push_back(cell);
	 cell.clear();
	 tbl->GetCell(cell,columnIndex2,listOut[i]);
	 _examplesDetail.push_back(cell);
	 cell.clear();
  }

}


void ItemLoadObj::LoadExamplesNDB() {
  int i;
  int errCode;
  int columnIndex1;
  int columnIndex2;
  string cell;
  vector<string> ColumnNames;
  vector<string> values;
  vector<unsigned int> listOut;
  ISTable* tbl;

  if (_name.empty() || _datablock.empty() || _category.empty() || _keyword.empty()) {
    _examplesCaseNDB.clear();
    _examplesDetailNDB.clear();
    return;
  }
  
  tbl = ci->GetTablePtr(_datablock, CIF_DDL_CATEGORY_NDB_ITEM_EXAMPLES);
  if (tbl == NULL) {
    _examplesCaseNDB.clear();
    _examplesDetailNDB.clear();
    return;
  }
  ColumnNames.push_back(CIF_DDL_ITEM_NAME);
  values.push_back(_name); 
    
  columnIndex1 = tbl->GetColumnIndex(CIF_DDL_ITEM_NAME);
  tbl->SetFlags(ISTable::DT_STRING | ISTable::CASE_INSENSE,columnIndex1);

  tbl->Search(values,ColumnNames,listOut,errCode);
  if (errCode<0){
    _examplesCaseNDB.clear();
    _examplesDetailNDB.clear();
    return;
  }
  
  columnIndex1 = tbl->GetColumnIndex(CIF_DDL_ITEM_CASE);
  columnIndex2 = tbl->GetColumnIndex(CIF_DDL_ITEM_DETAIL);
  for (i = 0; i < (int)listOut.size(); i++) {
	 tbl->GetCell(cell,columnIndex1,listOut[i]);
	 _examplesCaseNDB.push_back(cell);
	 cell.clear();
	 tbl->GetCell(cell,columnIndex2,listOut[i]);
	 _examplesDetailNDB.push_back(cell);
	 cell.clear();
  }

}



void ItemLoadObj::LoadEnumeration() {
  int i;
  ISTable * tbl;
  vector<string> ColumnNames;
  vector<string> values;
  vector<unsigned int> listOut;
  string cell;
  int errCode;
  int columnIndex1;
  int columnIndex2;

  if (_name.empty() || _datablock.empty() || _category.empty() || _keyword.empty()) {
    _enumeration.clear();
    _enumerationDetail.clear();
    return;
  }
  
  tbl = ci->GetTablePtr(_datablock, CIF_DDL_CATEGORY_ITEM_ENUMERATION);
  if (tbl == NULL) {
    _enumeration.clear();
    _enumerationDetail.clear();
    return;
  }

  ColumnNames.push_back(CIF_DDL_ITEM_NAME);
  values.push_back(_name); 
  
  columnIndex1 = tbl->GetColumnIndex(CIF_DDL_ITEM_NAME);
  tbl->SetFlags(ISTable::DT_STRING | ISTable::CASE_INSENSE,columnIndex1);

  tbl->Search(values,ColumnNames,listOut,errCode);
  if (errCode<0){
    _enumeration.clear();
    _enumerationDetail.clear();
    return;
  }

  columnIndex1 = tbl->GetColumnIndex(CIF_DDL_ITEM_VALUE);
  columnIndex2 = tbl->GetColumnIndex(CIF_DDL_ITEM_DETAIL);
  for (i = 0; i < (int)listOut.size(); i++) {
	 tbl->GetCell(cell,columnIndex1,listOut[i]);
	 _enumeration.push_back(cell);
	 cell.clear();
	 tbl->GetCell(cell,columnIndex2,listOut[i]);
	 _enumerationDetail.push_back(cell);
	 cell.clear();
  }

}


void ItemLoadObj::LoadEnumerationNDB() {
  int i;
  ISTable * tbl;
  vector<string> ColumnNames;
  vector<string> values;
  vector<unsigned int> listOut;
  string cell;
  int errCode;
  int columnIndex1;
  int columnIndex2;

  if (_name.empty() || _datablock.empty() || _category.empty() || _keyword.empty()) {
    _enumerationNDB.clear();
    _enumerationDetailNDB.clear();
    return;
  }
  
  tbl = ci->GetTablePtr(_datablock, CIF_DDL_CATEGORY_NDB_ITEM_ENUMERATION);
  if (tbl == NULL) {
    _enumerationNDB.clear();
    _enumerationDetailNDB.clear();
    return;
  }
  ColumnNames.push_back(CIF_DDL_ITEM_NAME);
  values.push_back(_name); 
  
  columnIndex1 = tbl->GetColumnIndex(CIF_DDL_ITEM_NAME);
  tbl->SetFlags(ISTable::DT_STRING | ISTable::CASE_INSENSE,columnIndex1);

  tbl->Search(values,ColumnNames,listOut,errCode);
  if (errCode<0){
    _enumerationNDB.clear();
    _enumerationDetailNDB.clear();
    return;
  }
  columnIndex1 = tbl->GetColumnIndex(CIF_DDL_ITEM_VALUE);
  columnIndex2 = tbl->GetColumnIndex(CIF_DDL_ITEM_DETAIL);
  for (i = 0; i < (int)listOut.size(); i++) {
	 tbl->GetCell(cell,columnIndex1,listOut[i]);
	 _enumerationNDB.push_back(cell);
	 cell.clear();
	 tbl->GetCell(cell,columnIndex2,listOut[i]);
	 _enumerationDetailNDB.push_back(cell);
	 cell.clear();
  }
  
}

void ItemLoadObj::LoadMandatoryCode() {
  if (_name.empty() || _datablock.empty() || _category.empty() || _keyword.empty()) {
    _mandatoryCode.clear();
    return;
  }
    _mandatoryCode = ci->GetItemMandatoryCodeInherit(_datablock, _category, _keyword, _decendency, _decendencyNum);
}

void ItemLoadObj::LoadDefaultValue() {
  if (_name.empty() || _datablock.empty() || _category.empty() || _keyword.empty()) {
    _defaultValue.clear();
    return;
  }
    _defaultValue = ci->GetItemDefaultValueInherit(_datablock, _category, _keyword, _decendency, _decendencyNum);
}

void ItemLoadObj::LoadUnits() {
  if (_name.empty() || _datablock.empty() || _category.empty() || _keyword.empty()) {
    _units.clear();
    return;
  }
    _units = ci->GetItemUnitsInherit(_datablock, _category, _keyword, _decendency, _decendencyNum);
}

void ItemLoadObj::LoadItemStructure() {
  ISTable * tbl;
  vector<string> ColumnNames;
  vector<string> values;
  vector<unsigned int> listOut;
  string cell;
  int errCode;
  int columnIndex1;
  int columnIndex2;

  if (_name.empty() || _datablock.empty() || _category.empty() || _keyword.empty()) {
    _itemStructure.clear();
    _itemStructureOrganization.clear();
    return;
  }

  tbl = ci->GetTablePtr(_datablock, CIF_DDL_CATEGORY_ITEM_STRUCTURE);
  if (tbl == NULL) {
    _itemStructure.clear();
    _itemStructureOrganization.clear();
    return;
  }
  ColumnNames.push_back(CIF_DDL_ITEM_NAME);
  values.push_back(_name); 
  
  columnIndex1 = tbl->GetColumnIndex(CIF_DDL_ITEM_NAME);
  tbl->SetFlags(ISTable::DT_STRING | ISTable::CASE_INSENSE,columnIndex1);

  tbl->Search(values,ColumnNames,listOut,errCode);
  if (errCode<0){
    _itemStructure.clear();
    _itemStructureOrganization.clear();
    return;
  }
  columnIndex1 = tbl->GetColumnIndex(CIF_DDL_ITEM_CODE);
  columnIndex2 = tbl->GetColumnIndex(CIF_DDL_ITEM_ORGANIZATION);
  tbl->GetCell(cell,columnIndex1,listOut[0]);
  _itemStructure = cell;
  cell.clear();
  tbl->GetCell(cell,columnIndex2,listOut[0]);
  _itemStructureOrganization = cell;

}

void ItemLoadObj::LoadRanges() {
  // int i;
  ISTable * tbl;
  vector<string> ColumnNames;
  vector<string> values;
  vector<unsigned int> listOut;
  string cell;
  int errCode;
  int columnIndex1;
  int columnIndex2;

  if (_name.empty() || _datablock.empty() || _category.empty() || _keyword.empty()) {
    _rangeMin.clear();
    _rangeMax.clear();
    return;
  }
  
  tbl = ci->GetTablePtr(_datablock, CIF_DDL_CATEGORY_ITEM_RANGE);
  if (tbl == NULL) {
    _rangeMin.clear();
    _rangeMax.clear();
    return;
  }
  ColumnNames.push_back(CIF_DDL_ITEM_NAME);
  values.push_back(_name); 
  
  columnIndex1 = tbl->GetColumnIndex(CIF_DDL_ITEM_NAME);
  tbl->SetFlags(ISTable::DT_STRING | ISTable::CASE_INSENSE,columnIndex1);

  tbl->Search(values,ColumnNames,listOut,errCode);
  if (errCode<0){
    _rangeMin.clear();
    _rangeMax.clear();
  }
  else {
	 columnIndex1 = tbl->GetColumnIndex(CIF_DDL_ITEM_MINIMUM);
	 columnIndex2 = tbl->GetColumnIndex(CIF_DDL_ITEM_MAXIMUM);
	 for (unsigned int i = 0; i < listOut.size(); i++) {
		tbl->GetCell(cell,columnIndex1,listOut[i]);
		_rangeMin.push_back(cell);
		cell.clear();
		tbl->GetCell(cell,columnIndex2,listOut[i]);
		_rangeMax.push_back(cell);
		cell.clear();
	 }
  }

}

void ItemLoadObj::LoadAliases() {
  int i;
  ISTable * tbl;
  vector<string> ColumnNames;
  vector<string> values;
  vector<unsigned int> listOut;
  string cell;
  int errCode;
  int columnIndex1;
  int columnIndex2;
  int columnIndex3;

  if (_name.empty() || _datablock.empty() || _category.empty() || _keyword.empty()) {
    _aliases.clear();
    _aliasesDictionary.clear();
    _aliasesDictionaryVersion.clear();
    return;
  }
  
  tbl = ci->GetTablePtr(_datablock, CIF_DDL_CATEGORY_ITEM_ALIASES);
  if (tbl == NULL) {
    _aliases.clear();
    _aliasesDictionary.clear();
    _aliasesDictionaryVersion.clear();
    return;
  }
  ColumnNames.push_back(CIF_DDL_ITEM_NAME);
  values.push_back(_name); 
  
  columnIndex1 = tbl->GetColumnIndex(CIF_DDL_ITEM_NAME);
  tbl->SetFlags(ISTable::DT_STRING | ISTable::CASE_INSENSE,columnIndex1);

  tbl->Search(values,ColumnNames,listOut,errCode);
  if (errCode<0){
    _aliases.clear();
    _aliasesDictionary.clear();
    _aliasesDictionaryVersion.clear();
  }
  else {
	 columnIndex1 = tbl->GetColumnIndex(CIF_DDL_ITEM_ALIAS_NAME);
	 columnIndex2 = tbl->GetColumnIndex(CIF_DDL_ITEM_DICTIONARY);
	 columnIndex3 = tbl->GetColumnIndex(CIF_DDL_ITEM_VERSION);
	 for (i = 0; i < (int)listOut.size(); i++) {
		tbl->GetCell(cell,columnIndex1,listOut[i]);
		_aliases.push_back(cell);
		cell.clear();
		tbl->GetCell(cell,columnIndex2,listOut[i]);
		_aliasesDictionary.push_back(cell);
		cell.clear();
		tbl->GetCell(cell,columnIndex3,listOut[i]);
		_aliasesDictionaryVersion.push_back(cell);
		cell.clear();
	 }
  }
}

void ItemLoadObj::LoadDependents() {
  if (_name.empty() || _datablock.empty() || _category.empty() || _keyword.empty()) {
    _dependents.clear();
    return;
  }
  ci->GetItemDependentsInherit(_dependents, _datablock, _category, _keyword, _decendency, _decendencyNum);
}

void ItemLoadObj::LoadRelated() {
  //int i;
  vector<string> ColumnNames;
  vector<string> values;
  vector<unsigned int> listOut;
  string cell;
  int errCode;
  int columnIndex1;
  int columnIndex2;

  if (_name.empty() || _datablock.empty() || _category.empty() || _keyword.empty()) {
    _related.clear();
    _relatedFunctionCode.clear();
    return;
  }

  ISTable * tbl = ci->GetTablePtr(_datablock, CIF_DDL_CATEGORY_ITEM_RELATED);
  if(tbl == NULL) {
    _related.clear();
    _relatedFunctionCode.clear();
    return;
  }
  ColumnNames.push_back(CIF_DDL_ITEM_NAME);
  values.push_back(_name);  

  columnIndex1 = tbl->GetColumnIndex(CIF_DDL_ITEM_NAME);
  tbl->SetFlags(ISTable::DT_STRING | ISTable::CASE_INSENSE,columnIndex1);

  tbl->Search(values,ColumnNames,listOut,errCode);
  if (errCode<0) {
    _related.clear();
    _relatedFunctionCode.clear();    
    return;
  }
  else {
	 columnIndex1 = tbl->GetColumnIndex(CIF_DDL_ITEM_RELATED_NAME);
	 columnIndex2 = tbl->GetColumnIndex(CIF_DDL_ITEM_FUNCTION_CODE);
	 for (unsigned int i = 0; i < listOut.size(); i++) {
		tbl->GetCell(cell,columnIndex1,listOut[i]);
		_related.push_back(cell);
		cell.clear();
		tbl->GetCell(cell,columnIndex2,listOut[i]);
		_relatedFunctionCode.push_back(cell);
		cell.clear();
	 }
  }

}

void ItemLoadObj::LoadSubcategories() {
  if (_name.empty() || _datablock.empty() || _category.empty() || _keyword.empty()) {
    _subcategories.clear();
    return;
  }
    ci->GetItemSubcategoriesInherit(_subcategories,
      _datablock, _category, _keyword, _decendency, _decendencyNum);
}

void ItemLoadObj::LoadLinkedChildren() {
  if (_name.empty() || _datablock.empty() || _category.empty() || _keyword.empty()) {
    _linkedChildren.clear();
    return;
  }
  ci->GetItemChildren(_linkedChildren, _datablock, _category, _keyword);
}

void ItemLoadObj::LoadLinkedParents() {
  if (_name.empty() || _datablock.empty() || _category.empty() || _keyword.empty()) {
    _linkedParents.clear();
    return;
  }
  ci->GetItemParents(_linkedParents, _datablock, _category, _keyword);
}

void ItemLoadObj::LoadMethods() {
  if (_name.empty() || _datablock.empty() || _category.empty() || _keyword.empty()) {
    _methods.clear();
    return;
  }

  ci->GetItemMethods(_methods, _datablock, _category, _keyword);
}

void ItemLoadObj::LoadTypeConditions() {
  if (_name.empty() || _datablock.empty() || _category.empty() || _keyword.empty()) {
    _typeConditions.clear();
    return;
  }
  ci->GetItemTypeConditions(_typeConditions, _datablock, _category, _keyword);
}

int ItemLoadObj::WriteObject() {
/*
------------------------------------------------------------------------------
  WriteObject() calls the private Write methods to write the whole object to
  persistent storage. It then tacks on an index into the file for purposes of
  reconstructing the object.
------------------------------------------------------------------------------
*/
  Word i;
  i = 0;
  if (_index) free(_index);
  _index = (Word *) malloc(ITEM_FIELDS*WORDSIZE);
  _index[0] = WriteDescription();
  _index[1] = WriteType();
  _index[2] = WritePrimitiveCode();
  _index[3] = WriteRegex();
  _index[4] = WriteExamplesCase();
  _index[5] = WriteExamplesDetail();
  _index[6] = WriteEnumeration();
  _index[7] = WriteEnumerationDetail();
  _index[8] = WriteMandatoryCode();
  _index[9] = WriteDefaultValue();
  _index[10] = WriteUnits();
  _index[11] = WriteRangeMin();
  _index[12] = WriteRangeMax();
  _index[13] = WriteAliases();
  _index[14] = WriteAliasesDictionary();
  _index[15] = WriteAliasesDictionaryVersion();
  _index[16] = WriteItemStructure();
  _index[17] = WriteItemStructureOrganization();
  _index[18] = WriteDependents();
  _index[19] = WriteRelated();
  _index[20] = WriteRelatedFunctionCode();
  _index[21] = WriteSubcategories();
  _index[22] = WriteLinkedChildren();
  _index[23] = WriteLinkedParents();
  _index[24] = WriteMethods();
  _index[25] = WriteTypeConditions();
  _index[26] = WriteDescriptionNDB();
  _index[27] = WriteExamplesCaseNDB();
  _index[28] = WriteExamplesDetailNDB();
  _index[29] = WriteEnumerationNDB();
  _index[30] = WriteEnumerationDetailNDB();
  int err = fnav->WriteWords(_index, ITEM_FIELDS, i);
  if (err == NO_ERROR) 
    return i; 
  else return err;
}

/*
  The following Write methods utilize the FileNavigator class to write the
  data of the object onto a persistent stroage file.
*/

int ItemLoadObj::WriteDescription() {
  Word i;
  i = 0;
  int err;
  if (_description.empty()) return NO_ERROR;
  if ((err = fnav->WriteString(_description, i))==NO_ERROR) return i;
  else return err; 
}

int ItemLoadObj::WriteDescriptionNDB() {
  Word i;
  i = 0;
  int err;
  if (_descriptionNDB.empty()) return NO_ERROR;
  if ((err = fnav->WriteString(_descriptionNDB, i))==NO_ERROR) return i;
  else return err; 
}

int ItemLoadObj::WriteType() {
  Word i;
  i = 0;
  int err;
  if (_type.empty()) return NO_ERROR;
  if ((err = fnav->WriteString(_type, i))==NO_ERROR) return i;
  else return err; 
}

int ItemLoadObj::WritePrimitiveCode() {
  Word i;
  i = 0;
  int err;
  if (_primitiveCode.empty()) return NO_ERROR;
  if ((err = fnav->WriteString(_primitiveCode, i))==NO_ERROR) return i;
  else return err; 
}

int ItemLoadObj::WriteRegex() {
  Word i;
  i = 0;
  int err;
  if (_regex.empty()) return NO_ERROR;
  if ((err = fnav->WriteString(_regex, i))==NO_ERROR) return i;
  else return err; 
}

int ItemLoadObj::WriteExamplesCase() {
  Word i;
  i = 0;
  int err;
  if (_examplesCase.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_examplesCase, i))
       == NO_ERROR) return i;
  else return err; 
}

int ItemLoadObj::WriteExamplesDetail() {
  Word i;
  i = 0;
  int err;
  if (_examplesDetail.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_examplesDetail, i))
       == NO_ERROR) return i;
  else return err; 
}

int ItemLoadObj::WriteEnumeration() {
  Word i;
  i = 0;
  int err;
  if (_enumeration.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_enumeration, i))
       == NO_ERROR) return i;
  else return err; 
}

int ItemLoadObj::WriteEnumerationDetail() {
  Word i;
  i = 0;
  int err;
  if (_enumerationDetail.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_enumerationDetail, i))
       == NO_ERROR) return i;
  else return err; 
}

int ItemLoadObj::WriteExamplesCaseNDB() {
  Word i;
  i = 0;
  int err;
  if (_examplesCaseNDB.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_examplesCaseNDB, i))
       == NO_ERROR) return i;
  else return err; 
}

int ItemLoadObj::WriteExamplesDetailNDB() {
  Word i;
  i = 0;
  int err;
  if (_examplesDetailNDB.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_examplesDetailNDB, i))
       == NO_ERROR) return i;
  else return err; 
}

int ItemLoadObj::WriteEnumerationNDB() {
  Word i;
  i = 0;
  int err;
  if (_enumerationNDB.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_enumerationNDB, i))
       == NO_ERROR) return i;
  else return err; 
}

int ItemLoadObj::WriteEnumerationDetailNDB() {
  Word i;
  i = 0;
  int err;
  if (_enumerationDetailNDB.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_enumerationDetailNDB, i))
       == NO_ERROR) return i;
  else return err; 
}

int ItemLoadObj::WriteMandatoryCode() {
  Word i;
  i = 0;
  int err;
  if (_mandatoryCode.empty()) return NO_ERROR;
  if ((err = fnav->WriteString(_mandatoryCode, i))==NO_ERROR) return i;
  else return err; 
}

int ItemLoadObj::WriteDefaultValue() {
  Word i;
  i = 0;
  int err;
  if (_defaultValue.empty()) return NO_ERROR;
  if ((err = fnav->WriteString(_defaultValue, i))==NO_ERROR) return i;
  else return err; 
}

int ItemLoadObj::WriteUnits() {
  Word i;
  i = 0;
  int err;
  if (_units.empty()) return NO_ERROR;
  if ((err = fnav->WriteString(_units, i))==NO_ERROR) return i;
  else return err; 
}

int ItemLoadObj::WriteRangeMin() {
  Word i;
  i = 0;
  int err;
  if (_rangeMin.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_rangeMin, i))==NO_ERROR) return i;
  else return err; 
}

int ItemLoadObj::WriteRangeMax() {
  Word i;
  i = 0;
  int err;
  if (_rangeMax.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_rangeMax, i))==NO_ERROR) return i;
  else return err; 
}

int ItemLoadObj::WriteAliases() {
  Word i;
  i = 0;
  int err;
  if (_aliases.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_aliases, i))==NO_ERROR) return i;
  else return err; 
}

int ItemLoadObj::WriteAliasesDictionary() {
  Word i;
  i = 0;
  int err;
  if (_aliasesDictionary.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_aliasesDictionary, i))==NO_ERROR) return i;
  else return err; 
}

int ItemLoadObj::WriteAliasesDictionaryVersion() {
  Word i;
  i = 0;
  int err;
  if (_aliasesDictionaryVersion.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_aliasesDictionaryVersion, i))==NO_ERROR) return i;
  else return err; 
}

int ItemLoadObj::WriteItemStructure() {
  Word i;
  i = 0;
  int err;
  if (_itemStructure.empty()) return NO_ERROR;
  if ((err = fnav->WriteString(_itemStructure, i))==NO_ERROR) return i;
  else return err; 
}

int ItemLoadObj::WriteItemStructureOrganization() {
  Word i;
  i = 0;
  int err;
  if (_itemStructureOrganization.empty()) return NO_ERROR;
  if ((err = fnav->WriteString(_itemStructureOrganization, i))==NO_ERROR) return i;
  else return err; 
}

int ItemLoadObj::WriteDependents() {
  Word i;
  i = 0;
  int err;
  if (_dependents.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_dependents, i))==NO_ERROR) return i;
  else return err; 
}

int ItemLoadObj::WriteRelated() {
  Word i;
  i = 0;
  int err;
  if (_related.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_related, i))==NO_ERROR) return i;
  else return err; 
}

int ItemLoadObj::WriteRelatedFunctionCode() {
  Word i;
  i = 0;
  int err;
  if (_relatedFunctionCode.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_relatedFunctionCode, i))==NO_ERROR) return i;
  else return err; 
}

int ItemLoadObj::WriteSubcategories() {
  Word i;
  i = 0;
  int err;
  if (_subcategories.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_subcategories, i))==NO_ERROR) return i;
  else return err; 
}

int ItemLoadObj::WriteLinkedChildren() {
  Word i;
  i = 0;
  int err;
  if (_linkedChildren.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_linkedChildren, i))==NO_ERROR) return i;
  else return err; 
}

int ItemLoadObj::WriteLinkedParents() {
  Word i;
  i = 0;
  int err;
  if (_linkedParents.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_linkedParents, i))==NO_ERROR) return i;
  else return err; 
}

int ItemLoadObj::WriteMethods() {
  Word i;
  i = 0;
  int err;
  if (_methods.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_methods, i))==NO_ERROR) return i;
  else return err; 
}

int ItemLoadObj::WriteTypeConditions() {
  Word i;
  i = 0;
  int err;
  if (_typeConditions.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_typeConditions, i))==NO_ERROR) return i;
  else return err; 
}

SubcategoryLoadObj::SubcategoryLoadObj() {
  Clear();
  if (!ci) ci = new CifInterface;
}

SubcategoryLoadObj::SubcategoryLoadObj(const string& dataBlockName,
                                                const string& subcategoryId) {
  if (!ci) ci = new CifInterface;
  Clear();
  SetId(dataBlockName, subcategoryId);
}  

void SubcategoryLoadObj::SetId(const string& dataBlockName,
                                            const string& subcategoryId) {

  _id = subcategoryId;  
  _datablock = dataBlockName;
}

int SubcategoryLoadObj::LoadObject() {
/*
-------------------------------------------------------------------------------
  LoadObject() will call all of the private Load methods so that all data is
  retrieved from ciflib and stored in the object. Returns KEYS_NOT_SET upon
  failure.
-------------------------------------------------------------------------------
*/
  if (_datablock.empty() || _id.empty()) return KEYS_NOT_SET; 
  LoadDescription();
  LoadExamples();
  LoadMethods();
  return NO_ERROR;
}

/*
   The following Load methods use ciflib (either directly or through the
   CifInterface class) to populate the data section of the object.
*/

void SubcategoryLoadObj::LoadDescription() {
  if (_id.empty() || _datablock.empty())
    _description.clear();
  else _description = ci->GetSubCategoryDescription(_datablock, _id);
}

void SubcategoryLoadObj::LoadExamples() {
  int i;
  ISTable * tbl;
  vector<string> ColumnNames;
  vector<string> values;
  vector<unsigned int> listOut;
  string cell;
  int errCode;
  int columnIndex1;
  int columnIndex2;

  if (_id.empty() || _datablock.empty()) {
    _examplesCase.clear();
    _examplesDetail.clear();
    return;
  }
  
  tbl = ci->GetTablePtr(_datablock, CIF_DDL_CATEGORY_SUB_CATEGORY_EXAMPLES);
  if (tbl == NULL) {
    _examplesCase.clear();
    _examplesDetail.clear();
    return;
  }
  ColumnNames.push_back(CIF_DDL_ITEM_NAME);
  values.push_back(_id); 
  
  columnIndex1 = tbl->GetColumnIndex(CIF_DDL_ITEM_NAME);
  tbl->SetFlags(ISTable::DT_STRING | ISTable::CASE_INSENSE,columnIndex1);

  tbl->Search(values,ColumnNames,listOut,errCode);
  if (errCode<0){
    _examplesCase.clear();
    _examplesDetail.clear();
    return;
  }

  columnIndex1 = tbl->GetColumnIndex(CIF_DDL_ITEM_CASE);
  columnIndex2 = tbl->GetColumnIndex(CIF_DDL_ITEM_DETAIL);

  for (i = 0; i < (int)listOut.size(); i++) {
	 tbl->GetCell(cell,columnIndex1,listOut[i]);
	 _examplesCase.push_back(cell);
	 cell.clear();
	 tbl->GetCell(cell,columnIndex2,listOut[i]);
	 _examplesDetail.push_back(cell);
	 cell.clear();
  }

}

void SubcategoryLoadObj::LoadMethods() {
  if (_id.empty() || _datablock.empty())
    _methods.clear();
  else
  {
    ci->GetSubCategoryMethods(_methods, _datablock, _id);
  }
}

int SubcategoryLoadObj::WriteObject() {
/*
------------------------------------------------------------------------------
  WriteObject() calls the private Write methods to write the whole object to
  persistent storage. It then tacks on an index into the file for purposes of
  reconstructing the object.
------------------------------------------------------------------------------
*/
  Word i;
  i = 0;
  if (_index) free(_index);
  _index = (Word *) malloc(SUBCATEGORY_FIELDS*WORDSIZE);
  _index[0] = WriteDescription();
  _index[1] = WriteExamplesCase();
  _index[2] = WriteExamplesDetail();
  _index[3] = WriteMethods();
  int err = fnav->WriteWords(_index, SUBCATEGORY_FIELDS, i);
  if (err == NO_ERROR) 
    return i; 
  else return err;
}

/*
  The following Write methods utilize the FileNavigator class to write the
  data of the object onto a persistent stroage file.
*/

int SubcategoryLoadObj::WriteDescription() {
  Word i;
  i = 0;
  int err;
  if (_description.empty()) return NO_ERROR;
  if ((err = fnav->WriteString(_description, i))==NO_ERROR) return i;
  else return err; 
}

int SubcategoryLoadObj::WriteExamplesCase() {
  Word i;
  i = 0;
  int err;
  if (_examplesCase.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_examplesCase, i))==NO_ERROR) return i;
  else return err; 
}

int SubcategoryLoadObj::WriteExamplesDetail() {
  Word i;
  i = 0;
  int err;
  if (_examplesDetail.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_examplesDetail, i))==NO_ERROR) return i;
  else return err; 
}

int SubcategoryLoadObj::WriteMethods() {
  Word i;
  i = 0;
  int err;
  if (_methods.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_methods, i))==NO_ERROR) return i;
  else return err; 
}



DictionaryLoadObj::DictionaryLoadObj() {
  Clear();
  _verbose = 0;
  if (!ci) ci = new CifInterface;
}

DictionaryLoadObj::DictionaryLoadObj(const string& d_name) {
  if (!ci) ci = new CifInterface;
  Clear();
  _verbose = 0;
  SetName(d_name.c_str());
}  

void DictionaryLoadObj::SetName(const string& d_name) {
  _name = d_name;
}

void DictionaryLoadObj::SetVerbose(int verbose) {
  _verbose = verbose;
}

int DictionaryLoadObj::LoadObject() {
/*
-------------------------------------------------------------------------------
  LoadObject() will call all of the private Load methods so that all data is
  retrieved from ciflib and stored in the object. Returns KEYS_NOT_SET upon
  failure.
-------------------------------------------------------------------------------
*/
//  cerr << "Dictionary LoadObject - Begin" << endl;
  if (_name.empty()) return KEYS_NOT_SET;
  LoadTitle();
  LoadVersion();
  LoadDescription();
  LoadHistory();
  LoadCategories();
  LoadSubcategories();
  LoadItems();
  LoadMethods();
  LoadMethodsList();
  LoadCategoryGroupList();
  LoadItemStructureList();
  LoadItemTypeList();
  LoadItemUnitsList();
  LoadItemUnitsConversion();
  return NO_ERROR;
}

/*
   The following Load methods use ciflib (either directly or through the
   CifInterface class) to populate the data section of the object.
*/

void DictionaryLoadObj::LoadTitle() {
  if (_name.empty()) _title.clear();
  else _title = ci->GetDictionaryTitle(_name);
}

void DictionaryLoadObj::LoadVersion() {
  if (_name.empty()) _version.clear();
  else _version = ci->GetDictionaryVersion(_name);
}

void DictionaryLoadObj::LoadDescription() {
  if (_name.empty()) _description.clear();
  else _description = ci->GetDictionaryDescription(_name);
}

void DictionaryLoadObj::LoadHistory() {
  int i;
  ISTable * tbl;
  string cell;
  int columnIndex1;
  int columnIndex2;
  int columnIndex3;

  if (_name.empty()) {
    _historyVersion.clear();
    _historyUpdate.clear();
    _historyRevision.clear();
    return;
  }    
  tbl = ci->GetTablePtr(_name, CIF_DDL_CATEGORY_DICTIONARY_HISTORY);
  if (tbl == NULL) {
    _historyVersion.clear();
    _historyUpdate.clear();
    _historyRevision.clear();
    return;
  }
  columnIndex1 = tbl->GetColumnIndex(CIF_DDL_ITEM_VERSION);
  columnIndex2 = tbl->GetColumnIndex(CIF_DDL_ITEM_UPDATE);
  columnIndex3 = tbl->GetColumnIndex(CIF_DDL_ITEM_REVISION);
  for (i=0; i<(int)tbl->GetNumRows(); i++) {
	 tbl->GetCell(cell,columnIndex1,i);
	 _historyVersion.push_back(cell);
	 cell.clear();
	 tbl->GetCell(cell,columnIndex2,i);
	 _historyUpdate.push_back(cell);
	 cell.clear();
	 tbl->GetCell(cell,columnIndex3,i);
	 _historyRevision.push_back(cell);
	 cell.clear();
  }
}  


void DictionaryLoadObj::LoadCategories() {
  if (_name.empty()) _categories.clear();
  else
    ci->GetDictionaryCategories(_categories, _name);
}

void DictionaryLoadObj::LoadSubcategories() {
  if (_name.empty())
    _subcategories.clear();
  else
    ci->GetDictionarySubcategories(_subcategories, _name);
}

void DictionaryLoadObj::LoadItems() {
  if (_name.empty()) _items.clear();
  else
    ci->GetDictionaryItems(_items, _name);
}

void DictionaryLoadObj::LoadMethods() {
  int i;
  ISTable * tbl;
  string cell;
  int columnIndex1;
  int columnIndex2;

  if (_name.empty()) {
	 _methods.clear();
	 return;
  }
  tbl = ci->GetTablePtr(_name, CIF_DDL_CATEGORY_DATABLOCK_METHODS);
  if (tbl == NULL) {
	 _methods.clear();
    return;
  }
  columnIndex1 = tbl->GetColumnIndex(CIF_DDL_ITEM_DATABLOCK_ID);
  columnIndex2 = tbl->GetColumnIndex(CIF_DDL_ITEM_METHOD_ID);

  for (i=0; i<(int)tbl->GetNumRows(); i++) {
	 tbl->GetCell(cell,columnIndex1,i);
	 _methods.push_back(cell);
	 cell.clear();
	 tbl->GetCell(cell,columnIndex2,i);
	 _methodsId.push_back(cell);
	 cell.clear();
  }
}

void DictionaryLoadObj::LoadMethodsList() {
  int i;
  ISTable * tbl;
  string cell;
  int columnIndex1;
  int columnIndex2;
  int columnIndex3;
  int columnIndex4;
  int columnIndex5;
  
  if (_name.empty()) {
    _methodsList.clear();
    _methodsListDetail.clear();
    _methodsListInline.clear();
    _methodsListCode.clear();
    _methodsListLanguage.clear();
    return;
  }
  tbl = ci->GetTablePtr(_name, CIF_DDL_CATEGORY_METHOD_LIST);
  if (tbl == NULL) {
    _methodsList.clear();
    _methodsListDetail.clear();
    _methodsListInline.clear();
    _methodsListCode.clear();
    _methodsListLanguage.clear();
    return;
  }
  columnIndex1 = tbl->GetColumnIndex(CIF_DDL_ITEM_ID);
  columnIndex2 = tbl->GetColumnIndex(CIF_DDL_ITEM_DETAIL);
  columnIndex3 = tbl->GetColumnIndex(CIF_DDL_ITEM_INLINE);
  columnIndex4 = tbl->GetColumnIndex(CIF_DDL_ITEM_CODE);
  columnIndex5 = tbl->GetColumnIndex(CIF_DDL_ITEM_LANGUAGE);

  for (i=0; i<(int)tbl->GetNumRows(); i++) {
	 tbl->GetCell(cell,columnIndex1,i);
	 _methodsList.push_back(cell);
	 cell.clear();
	 tbl->GetCell(cell,columnIndex2,i);
	 _methodsListDetail.push_back(cell);
	 cell.clear();
	 tbl->GetCell(cell,columnIndex3,i);
	 _methodsListInline.push_back(cell);
	 cell.clear();
	 tbl->GetCell(cell,columnIndex4,i);
	 _methodsListCode.push_back(cell);
	 cell.clear();
	 tbl->GetCell(cell,columnIndex5,i);
	 _methodsListLanguage.push_back(cell);
	 cell.clear();
  }
}

void DictionaryLoadObj::LoadCategoryGroupList() {
  int i;
  ISTable * tbl;
  string cell;
  int columnIndex1;
  int columnIndex2;
  int columnIndex3;

  if (_name.empty()) {
    _categoryGroupList.clear();
    _categoryGroupListParents.clear();
    _categoryGroupListDescription.clear();
    return;
  }
  tbl = ci->GetTablePtr(_name, CIF_DDL_CATEGORY_CATEGORY_GROUP_LIST);
  if (tbl == NULL) {
    _categoryGroupList.clear();
    _categoryGroupListParents.clear();
    _categoryGroupListDescription.clear();
    return;
  }
  columnIndex1 = tbl->GetColumnIndex(CIF_DDL_ITEM_ID);
  columnIndex2 = tbl->GetColumnIndex(CIF_DDL_ITEM_PARENT_ID);
  columnIndex3 = tbl->GetColumnIndex(CIF_DDL_ITEM_DESCRIPTION);
  
  for (i=0; i<(int)tbl->GetNumRows(); i++) {
	 tbl->GetCell(cell,columnIndex1,i);
	 _categoryGroupList.push_back(cell);
	 cell.clear();
	 tbl->GetCell(cell,columnIndex2,i);
	 _categoryGroupListParents.push_back(cell);
	 cell.clear();
	 tbl->GetCell(cell,columnIndex3,i);
	 _categoryGroupListDescription.push_back(cell);
	 cell.clear();

  }
}

void DictionaryLoadObj::LoadItemStructureList() {
  int i;
  ISTable * tbl;
  string cell;
  int columnIndex1;
  int columnIndex2;
  int columnIndex3;

  if (_name.empty()) {
    _itemStructureListCode.clear();
    _itemStructureListIndex.clear();
    _itemStructureListDimension.clear();
    return;
  }
  tbl = ci->GetTablePtr(_name, CIF_DDL_CATEGORY_ITEM_STRUCTURE_LIST);
  if (tbl == NULL) {
    _itemStructureListCode.clear();
    _itemStructureListIndex.clear();
    _itemStructureListDimension.clear();
    return;
  }
  columnIndex1 = tbl->GetColumnIndex(CIF_DDL_ITEM_CODE);
  columnIndex2 = tbl->GetColumnIndex(CIF_DDL_ITEM_INDEX);
  columnIndex3 = tbl->GetColumnIndex(CIF_DDL_ITEM_DIMENSION);

  for (i=0; i<(int)tbl->GetNumRows(); i++) {
	 tbl->GetCell(cell,columnIndex1,i);
	 _itemStructureListCode.push_back(cell);
	 cell.clear();
	 tbl->GetCell(cell,columnIndex2,i);
	 _itemStructureListIndex.push_back(cell);
	 cell.clear();
	 tbl->GetCell(cell,columnIndex3,i);
	 _itemStructureListDimension.push_back(cell);
	 cell.clear();
  }
}

void DictionaryLoadObj::LoadItemTypeList() {
  int i;
  ISTable * tbl;
  string cell;
  int columnIndex1;
  int columnIndex2;
  int columnIndex3;
  int columnIndex4;


  if (_name.empty()) {
    _itemTypeListCode.clear();
    _itemTypeListPrimitiveCode.clear();
    _itemTypeListConstruct.clear();
    _itemTypeListDetail.clear();
    return;
  }
  tbl = ci->GetTablePtr(_name, CIF_DDL_CATEGORY_ITEM_TYPE_LIST);
  if (tbl == NULL) {
    _itemTypeListCode.clear();
    _itemTypeListPrimitiveCode.clear();
    _itemTypeListConstruct.clear();
    _itemTypeListDetail.clear();
    return;
  }
  columnIndex1 = tbl->GetColumnIndex(CIF_DDL_ITEM_CODE);
  columnIndex2 = tbl->GetColumnIndex(CIF_DDL_ITEM_PRIMITIVE_CODE);
  columnIndex3 = tbl->GetColumnIndex(CIF_DDL_ITEM_CONSTRUCT);
  columnIndex4 = tbl->GetColumnIndex(CIF_DDL_ITEM_DETAIL);
  
  for (i=0; i<(int)tbl->GetNumRows(); i++) {
	 tbl->GetCell(cell,columnIndex1,i);
	 _itemTypeListCode.push_back(cell);
	 cell.clear();
	 tbl->GetCell(cell,columnIndex2,i);
	 _itemTypeListPrimitiveCode.push_back(cell);
	 cell.clear();
	 tbl->GetCell(cell,columnIndex3,i);
	 _itemTypeListConstruct.push_back(cell);
	 cell.clear();
	 tbl->GetCell(cell,columnIndex4,i);
	 _itemTypeListDetail.push_back(cell);
	 cell.clear();
  }
}

void DictionaryLoadObj::LoadItemUnitsList() {
  int i;
  ISTable * tbl;
  string cell;
  int columnIndex1;
  int columnIndex2;


  if (_name.empty()) {
    _itemUnitsListCode.clear();
    _itemUnitsListDetail.clear();
    return;
  }
  tbl = ci->GetTablePtr(_name, CIF_DDL_CATEGORY_ITEM_UNITS_LIST);
  if (tbl == NULL) {
    _itemUnitsListCode.clear();
    _itemUnitsListDetail.clear();
    return;
  }
  columnIndex1 = tbl->GetColumnIndex(CIF_DDL_ITEM_CODE);
  columnIndex2 = tbl->GetColumnIndex(CIF_DDL_ITEM_DETAIL);
  
  for (i=0; i<(int)tbl->GetNumRows(); i++) {
	 tbl->GetCell(cell,columnIndex1,i);
	 _itemUnitsListCode.push_back(cell);
	 cell.clear();
	 tbl->GetCell(cell,columnIndex2,i);
	 _itemUnitsListDetail.push_back(cell);
	 cell.clear();
  }
}

void DictionaryLoadObj::LoadItemUnitsConversion() {
  int i;
  ISTable * tbl;
  string cell;
  int columnIndex1;
  int columnIndex2;
  int columnIndex3;
  int columnIndex4;

  if (_name.empty()) {
    _itemUnitsConversionOperator.clear();
    _itemUnitsConversionFactor.clear();
    _itemUnitsConversionFromCode.clear();
    _itemUnitsConversionToCode.clear();
    return;
  }
  tbl = ci->GetTablePtr(_name,CIF_DDL_CATEGORY_ITEM_UNITS_CONVERSION);
  if (tbl == NULL) {
    _itemUnitsConversionOperator.clear();
    _itemUnitsConversionFactor.clear();
    _itemUnitsConversionFromCode.clear();
    _itemUnitsConversionToCode.clear();
    return;
  }
  columnIndex1 = tbl->GetColumnIndex(CIF_DDL_ITEM_OPERATOR);
  columnIndex2 = tbl->GetColumnIndex(CIF_DDL_ITEM_FACTOR);
  columnIndex3 = tbl->GetColumnIndex(CIF_DDL_ITEM_FROM_CODE);
  columnIndex4 = tbl->GetColumnIndex(CIF_DDL_ITEM_TO_CODE);

  for (i=0; i<(int)tbl->GetNumRows(); i++) {
	 tbl->GetCell(cell,columnIndex1,i);
	 _itemUnitsConversionOperator.push_back(cell);
	 cell.clear();
	 tbl->GetCell(cell,columnIndex2,i);
	 _itemUnitsConversionFactor.push_back(cell);
	 cell.clear();
	 tbl->GetCell(cell,columnIndex3,i);
	 _itemUnitsConversionFromCode.push_back(cell);
	 cell.clear();
	 tbl->GetCell(cell,columnIndex4,i);
	 _itemUnitsConversionToCode.push_back(cell);
	 cell.clear();
  }
}

int DictionaryLoadObj::WriteObject() {
/*
------------------------------------------------------------------------------
  WriteObject() calls the private Write methods to write the whole object to
  persistent storage. It then tacks on an index into the file for purposes of
  reconstructing the object.
------------------------------------------------------------------------------
*/
  Word i;
  i = 0;
  if (_index) free(_index);
  _index = (Word *) malloc(DICTIONARY_FIELDS*WORDSIZE);

  ItemLoadObj * ilo;
  SubcategoryLoadObj * slo;
  CategoryLoadObj * clo;
  if (_items.empty())
    _itemLocations=NULL;
  else
    _itemLocations = (uWord *) calloc(_items.size(), WORDSIZE);

  if (_subcategories.empty())
    _subcategoryLocations=NULL;
  else
  _subcategoryLocations = (uWord *) calloc(_subcategories.size(), WORDSIZE);

  if (_categories.empty())
    _categoryLocations=NULL;
  else
  _categoryLocations = (uWord *) calloc(_categories.size(), WORDSIZE);

  for (i = 0; i < (int)_items.size(); i++) {
    ilo = new ItemLoadObj(_name, _items[i].c_str());
    ilo->LoadObject();
    if (_verbose) ilo->PrintObject();
    _itemLocations[i] = ilo->WriteObject();
    delete ilo;
  }

  for (i = 0; i < (int)_subcategories.size(); i++) {
    slo = new SubcategoryLoadObj(_name, _subcategories[i].c_str());
    slo->LoadObject();
    if (_verbose) slo->PrintObject();
    _subcategoryLocations[i] = slo->WriteObject();
    delete slo;
  }

  for (i = 0; i < (int)_categories.size(); i++) {
    clo = new CategoryLoadObj(_name, _categories[i].c_str());
    clo->LoadObject();
    if (_verbose) clo->PrintObject();
    _categoryLocations[i] = clo->WriteObject();
    delete clo;
  }


  _index[0] = WriteTitle();
  _index[1] = WriteVersion();
  _index[2] = WriteDescription();
  _index[3] = WriteHistoryVersion();
  _index[4] = WriteHistoryUpdate();
  _index[5] = WriteHistoryRevision();


  _index[6] = WriteCategories();
  _index[7] = WriteSubcategories();
  _index[8] = WriteItems();
  _index[9] = WriteMethods();
  _index[10] = WriteMethodsId();
  _index[11] = WriteMethodsList();
  _index[12] = WriteMethodsListDetail();
  _index[13] = WriteMethodsListInline();
  _index[14] = WriteMethodsListCode();
  _index[15] = WriteMethodsListLanguage();


  _index[16] = WriteCategoryGroupList();
  _index[17] = WriteCategoryGroupListParents();
  _index[18] = WriteCategoryGroupListDescription();
  _index[19] = WriteItemStructureListCode();
  _index[20] = WriteItemStructureListIndex();
  _index[21] = WriteItemStructureListDimension();


  _index[22] = WriteItemTypeListCode();
  _index[23] = WriteItemTypeListPrimitiveCode();
  _index[24] = WriteItemTypeListConstruct();
  _index[25] = WriteItemTypeListDetail();


  _index[26] = WriteItemUnitsListCode();
  _index[27] = WriteItemUnitsListDetail();
  _index[28] = WriteItemUnitsConversionOperator();
  _index[29] = WriteItemUnitsConversionFactor();
  _index[30] = WriteItemUnitsConversionFromCode();
  _index[31] = WriteItemUnitsConversionToCode();


  _index[32] = WriteItemLocations();
  _index[33] = WriteSubcategoryLocations();
  _index[34] = WriteCategoryLocations();

  int err = fnav->WriteWords(_index, DICTIONARY_FIELDS, i);
  if (err == NO_ERROR) 
    return i; 
  else return err;
}

/*
  The following Write methods utilize the FileNavigator class to write the
  data of the object onto a persistent stroage file.
*/

int DictionaryLoadObj::WriteTitle() {
  Word i;
  i = 0;
  int err;
  if (_title.empty()) return NO_ERROR;
  if ((err = fnav->WriteString(_title, i))==NO_ERROR) return i;
  else return err; 
}

int DictionaryLoadObj::WriteVersion() {
  Word i;
  i = 0;
  int err;
  if (_version.empty()) return NO_ERROR;
  if ((err = fnav->WriteString(_version, i))==NO_ERROR) return i;
  else return err; 
}

int DictionaryLoadObj::WriteDescription() {
  Word i;
  i = 0;
  int err;
  if (_description.empty()) return NO_ERROR;
  if ((err = fnav->WriteString(_description, i))==NO_ERROR) return i;
  else return err; 
}

int DictionaryLoadObj::WriteHistoryVersion() {
  Word i;
  i = 0;
  int err;
  if (_historyVersion.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_historyVersion, i))==NO_ERROR) return i;
  else return err; 
}

int DictionaryLoadObj::WriteHistoryUpdate() {
  Word i;
  i = 0;
  int err;
  if (_historyUpdate.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_historyUpdate, i))==NO_ERROR) return i;
  else return err; 
}

int DictionaryLoadObj::WriteHistoryRevision() {
  Word i;
  i = 0;
  int err;
  if (_historyRevision.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_historyRevision, i))==NO_ERROR) return i;
  else return err; 
}

int DictionaryLoadObj::WriteCategories() {
  Word i;
  i = 0;
  int err;
  if (_categories.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_categories, i))==NO_ERROR) return i;
  else return err; 
}

int DictionaryLoadObj::WriteSubcategories() {
  Word i;
  i = 0;
  int err;
  if (_subcategories.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_subcategories, i))==NO_ERROR) return i;
  else return err; 
}

int DictionaryLoadObj::WriteItems() {
  Word i;
  i = 0;
  int err;
  if (_items.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_items, i))==NO_ERROR) return i;
  else return err; 
}

int DictionaryLoadObj::WriteMethods() {
  Word i;
  i = 0;
  int err;
  if (_methods.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_methods, i))==NO_ERROR) return i;
  else return err; 
}

int DictionaryLoadObj::WriteMethodsId() {
  Word i;
  i = 0;
  int err;
  if (_methodsId.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_methodsId, i))==NO_ERROR) return i;
  else return err; 
}

int DictionaryLoadObj::WriteMethodsList() {
  Word i;
  i = 0;
  int err;
  if (_methodsList.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_methodsList, i))==NO_ERROR) return i;
  else return err; 
}

int DictionaryLoadObj::WriteMethodsListDetail() {
  Word i;
  i = 0;
  int err;
  if (_methodsListDetail.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_methodsListDetail, i))==NO_ERROR) return i;
  else return err; 
}

int DictionaryLoadObj::WriteMethodsListInline() {
  Word i;
  i = 0;
  int err;
  if (_methodsListInline.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_methodsListInline, i))==NO_ERROR) return i;
  else return err; 
}

int DictionaryLoadObj::WriteMethodsListCode() {
  Word i;
  i = 0;
  int err;
  if (_methodsListCode.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_methodsListCode, i))==NO_ERROR) return i;
  else return err; 
}

int DictionaryLoadObj::WriteMethodsListLanguage() {
  Word i;
  i = 0;
  int err;
  if (_methodsListLanguage.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_methodsListLanguage, i))==NO_ERROR) return i;
  else return err; 
}

int DictionaryLoadObj::WriteCategoryGroupList() {
  Word i;
  i = 0;
  int err;
  if (_categoryGroupList.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_categoryGroupList, i))==NO_ERROR) return i;
  else return err; 
}

int DictionaryLoadObj::WriteCategoryGroupListParents() {
  Word i;
  i = 0;
  int err;
  if (_categoryGroupListParents.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_categoryGroupListParents, i))==NO_ERROR) return i;
  else return err; 
}

int DictionaryLoadObj::WriteCategoryGroupListDescription() {
  Word i;
  i = 0;
  int err;
  if (_categoryGroupListDescription.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_categoryGroupListDescription, i))==NO_ERROR) return i;
  else return err; 
}

int DictionaryLoadObj::WriteItemStructureListCode() {
  Word i;
  i = 0;
  int err;
  if (_itemStructureListCode.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_itemStructureListCode, i))==NO_ERROR) return i;
  else return err; 
}

int DictionaryLoadObj::WriteItemStructureListIndex() {
  Word i;
  i = 0;
  int err;
  if (_itemStructureListIndex.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_itemStructureListIndex, i))==NO_ERROR) return i;
  else return err; 
}

int DictionaryLoadObj::WriteItemStructureListDimension() {
  Word i;
  i = 0;
  int err;
  if (_itemStructureListDimension.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_itemStructureListDimension, i))==NO_ERROR) return i;
  else return err; 
}

int DictionaryLoadObj::WriteItemTypeListCode() {
  Word i;
  i = 0;
  int err;
  if (_itemTypeListCode.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_itemTypeListCode, i)) == NO_ERROR) return i;
  else return err; 
}

int DictionaryLoadObj::WriteItemTypeListPrimitiveCode() {
  Word i;
  i = 0;
  int err;
  if (_itemTypeListPrimitiveCode.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_itemTypeListPrimitiveCode, i)) == NO_ERROR) return i;
  else return err; 
}

int DictionaryLoadObj::WriteItemTypeListConstruct() {
  Word i;
  i = 0;
  int err;
  if (_itemTypeListConstruct.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_itemTypeListConstruct, i))==NO_ERROR) return i;
  else return err; 
}

int DictionaryLoadObj::WriteItemTypeListDetail() {
  Word i;
  i = 0;
  int err;
  if (_itemTypeListDetail.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_itemTypeListDetail, i)) == NO_ERROR) return i;
  else return err; 
}

int DictionaryLoadObj::WriteItemUnitsListCode() {
  Word i;
  i = 0;
  int err;
  if (_itemUnitsListCode.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_itemUnitsListCode, i)) == NO_ERROR) return i;
  else return err; 
}

int DictionaryLoadObj::WriteItemUnitsListDetail() {
  Word i;
  i = 0;
  int err;
  if (_itemUnitsListDetail.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_itemUnitsListDetail, i))==NO_ERROR) return i;
  else return err; 
}

int DictionaryLoadObj::WriteItemUnitsConversionOperator() {
  Word i;
  i = 0;
  int err;
  if (_itemUnitsConversionOperator.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_itemUnitsConversionOperator, i))==NO_ERROR) return i;
  else return err; 
}

int DictionaryLoadObj::WriteItemUnitsConversionFactor() {
  Word i;
  i = 0;
  int err;
  if (_itemUnitsConversionFactor.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_itemUnitsConversionFactor, i))==NO_ERROR) return i;
  else return err; 
}

int DictionaryLoadObj::WriteItemUnitsConversionFromCode() {
  Word i;
  i = 0;
  int err;
  if (_itemUnitsConversionFromCode.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_itemUnitsConversionFromCode, i))==NO_ERROR) return i;
  else return err; 
}

int DictionaryLoadObj::WriteItemUnitsConversionToCode() {
  Word i;
  i = 0;
  int err;
  if (_itemUnitsConversionToCode.empty()) return NO_ERROR;
  if ((err = fnav->WriteStrings(_itemUnitsConversionToCode, i))==NO_ERROR) return i;
  else return err; 
}

int DictionaryLoadObj::WriteItemLocations() {
  Word i;
  i = 0;
  int err;
  if (!_itemLocations) return NO_ERROR;
  if ((err = fnav->WriteUWords(_itemLocations, _items.size(), i)) ==NO_ERROR) return i;
  else return err;
}

int DictionaryLoadObj::WriteSubcategoryLocations() {
  Word i;
  i = 0;
  int err;
  if (!_subcategoryLocations) return NO_ERROR;
  if ((err = fnav->WriteUWords(_subcategoryLocations, _subcategories.size(), i)) == NO_ERROR) return i;
  else return err;
}

int DictionaryLoadObj::WriteCategoryLocations() {
  Word i;
  i = 0;
  int err;
  if (!_categoryLocations) return NO_ERROR;
  if ((err = fnav->WriteUWords(_categoryLocations, _categories.size(), i)) == NO_ERROR) return i;
  else return err;
}

void DictionaryLoadObj::Reset() {
/*
-------------------------------------------------------------------------------
  Reset() frees owned memory and reinitializes the object
-------------------------------------------------------------------------------
*/
  if (_itemLocations) free(_itemLocations);
  if (_subcategoryLocations) free(_subcategoryLocations);
  if (_categoryLocations) free(_categoryLocations);
  DictionaryObjBase::Reset();
}

void DictionaryLoadObj::Clear() {
/*
-------------------------------------------------------------------------------
  Clear() initializes the object
-------------------------------------------------------------------------------
*/
  DictionaryObjBase::Clear();
  _itemLocations = NULL;
  _subcategoryLocations = NULL;
  _categoryLocations = NULL;
}

